import logging

from pprint import pprint as pp
LOG = logging.getLogger(__name__)

def format_info(variant, variant_type='snv'):
    """Format the info field for SNV variants
    
    Args:
        variant(dict)
        variant_type(str): snv or sv
    
    Returns:
        vcf_info(str): A VCF formated info field
    """
    
    observations = variant.get('observations',0)

    homozygotes = variant.get('homozygote')
    hemizygotes = variant.get('hemizygote')

    
    vcf_info = f"Obs={observations}"
    if homozygotes:
        vcf_info += f";Hom={homozygotes}"
    if hemizygotes:
        vcf_info += f";Hem={hemizygotes}"

    # This is SV specific
    if variant_type == 'sv':
        end = int((variant['end_left'] + variant['end_right'])/2)
        
        vcf_info += f";SVTYPE={variant['sv_type']};END={end};SVLEN={variant['length']}"

    return vcf_info

def format_variant(variant, variant_type='snv'):
    """Convert variant information to a VCF formated string
    
    Args:
        variant(dict)
        variant_type(str)
    
    Returns:
        vcf_variant(str)
    """
    chrom = variant.get('chrom')
    pos = variant.get('start')
    
    ref = variant.get('ref')
    alt = variant.get('alt')
    pp(variant)
    
    if variant_type == 'sv':
        pos = int((variant['pos_left'] + variant['pos_right'])/2)
        ref = 'N'
        alt = f"<{variant['sv_type']}>"

    info = None
    
    info = format_info(variant, variant_type=variant_type)

    variant_line = f"{chrom}\t{pos}\t.\t{ref}\t{alt}\t.\t.\t{info}"
    
    return variant_line


import logging
from pymongo import MongoClient

from loqusdb.plugins import Base
from mongo_adapter import MongoAdapter as BaseAdapter

from . import VariantMixin, CaseMixin

from loqusdb import INDEXES

logger = logging.getLogger(__name__)


class MongoAdapter(BaseAdapter, VariantMixin, CaseMixin, Base):
    """docstring for MongoAdapter"""

    def wipe_db(self):
        """Wipe the whole database"""
        logger.warning("Wiping the whole database")
        self.client.drop_database(self.db_name)
        logger.debug("Database wiped")

    def indexes(self, collection=None):
         """Return a list with the current indexes

         Skip the mandatory _id_ indexes

         Args:
             collection(str)

         Returns:
             indexes(list)
         """
         indexes = []
         for collection_name in self.db.list_collection_names():
             if collection and collection != collection_name:
                 continue
             for index_name in self.db[collection_name].index_information():
                 if index_name != '_id_':
                     indexes.append(index_name)
         return indexes

    def check_indexes(self):
        """Check if the indexes exists"""
        for collection_name in INDEXES:
            existing_indexes = self.indexes(collection_name)
            indexes = INDEXES[collection_name]
            for index in indexes:
                index_name = index.document.get('name')
                if not index_name in existing_indexes:
                    logger.warning("Index {0} missing. Run command `loqusdb index`".format(index_name))
                    return
        logger.info("All indexes exists")

    def ensure_indexes(self):
        """Update the indexes"""
        for collection_name in INDEXES:
            existing_indexes = self.indexes(collection_name)
            indexes = INDEXES[collection_name]
            for index in indexes:
                index_name = index.document.get('name')
                if index_name in existing_indexes:
                    logger.debug("Index exists: %s" % index_name)
                    self.db[collection_name].drop_index(index_name)
            logger.info("creating indexes for collection {0}: {1}".format(
                collection_name,
                ', '.join([index.document.get('name') for index in indexes]),
                )
            )
            self.db[collection_name].create_indexes(indexes)



import pytest
from loqusdb.utils.case import update_case
from loqusdb.exceptions import CaseError

def test_update_case(case_obj, sv_case_obj):
    ## GIVEN a case with snv info and a case with sv info
    assert sv_case_obj['vcf_path'] is None
    assert case_obj['vcf_path'] is not None
    assert case_obj['vcf_sv_path'] is None
    assert sv_case_obj['vcf_sv_path'] is not None
    
    ## WHEN updating the case
    updated_case = update_case(case_obj=sv_case_obj, existing_case=case_obj)
    
    ## THEN assert that the cases have been merged without affecting the original cases
    assert updated_case['vcf_path'] is not None
    assert updated_case['vcf_path'] == case_obj['vcf_path']
    assert sv_case_obj['vcf_path'] is None

    assert updated_case['vcf_sv_path'] is not None
    assert updated_case['vcf_sv_path'] == sv_case_obj['vcf_sv_path']
    assert case_obj['vcf_sv_path'] is None
    
    assert updated_case['nr_variants'] == case_obj['nr_variants']
    assert updated_case['nr_sv_variants'] == sv_case_obj['nr_sv_variants']

def test_update_existing_vcf(case_obj):
    ## GIVEN a case object with VCF information
    ## WHEN trying to update existing VCF information
    ## THEN assert that a CaseError is raised    
    with pytest.raises(CaseError):
        updated_case = update_case(case_obj, case_obj)



import pytest

from loqusdb.utils.load import load_database
from loqusdb.exceptions import CaseError

def test_load_database(vcf_path, ped_path, real_mongo_adapter, case_id):
    mongo_adapter = real_mongo_adapter
    db = mongo_adapter.db

    load_database(
        adapter=mongo_adapter,
        variant_file=vcf_path,
        family_file=ped_path,
        family_type='ped',
    )

    mongo_case = db.case.find_one()

    assert mongo_case['case_id'] == case_id

def test_load_database_alternative_ped(vcf_path, ped_path, real_mongo_adapter, case_id):
    mongo_adapter = real_mongo_adapter
    db = mongo_adapter.db

    load_database(
        adapter=mongo_adapter,
        variant_file=vcf_path,
        family_file=ped_path,
        family_type='ped',
        case_id='alternative'
    )

    mongo_case = db.case.find_one()
    mongo_variant = db.variant.find_one()

    assert mongo_case['case_id'] == 'alternative'
    assert mongo_variant['families'] == ['alternative']

def test_load_database_wrong_ped(vcf_path, funny_ped_path, real_mongo_adapter):
    mongo_adapter = real_mongo_adapter
    ## GIVEN a vcf and ped file with wrong individuals
    ## WHEN loading the information
    ## THEN Error should be raised since individuals is not in vcf
    with pytest.raises(CaseError):
        load_database(
            adapter=mongo_adapter,
            variant_file=vcf_path,
            family_file=funny_ped_path,
            family_type='ped',
        )



import logging
import click

from . import base_command

LOG = logging.getLogger(__name__)

@base_command.command('identity', short_help="Search identity collection")
@click.option('-v', '--variant-id',
    help='Search for a variant ID.'
)
@click.pass_context
def identity(ctx, variant_id):
    """Check how well SVs are working in the database
        
    """
    if not variant_id:
        LOG.warning("Please provide a variant id")
        ctx.abort()

    adapter = ctx.obj['adapter']
    version = ctx.obj['version']
    
    LOG.info("Search variants {0}".format(adapter))
    
    result = adapter.get_clusters(variant_id)
    if result.count() == 0:
        LOG.info("No hits for variant %s", variant_id)
        return
    
    for res in result:
        click.echo(res)


import pytest

from loqusdb.utils.vcf import (check_vcf, get_vcf)
from loqusdb.exceptions import VcfError

def test_check_vcf_correct(vcf_path):
    ## GIVEN a vcf file and a counter that checks the number of variants
    true_nr = 0
    
    with open(vcf_path, 'r') as f:
        for line in f:
            if not line.startswith('#'):
                true_nr += 1
    
    ## WHEN collecting the VCF info
    vcf_info = check_vcf(vcf_path)
    
    ## THEN assert that the number of variants collected is correct
    assert vcf_info['nr_variants'] == true_nr
    ## THEN assert that the variant type is correct
    assert vcf_info['variant_type'] == 'snv'

def test_check_vcf_double_variant(double_vcf_path):
    ## GIVEN a variant file where a variant is duplicated
    ## WHEN checking the vcf
    ## THEN assert that the function raises a VcfError
    with pytest.raises(VcfError):
        check_vcf(double_vcf_path)

def test_check_vcf_unsorted(unsorted_vcf_path):
    ## GIVEN a vcf file with unsorted variants
    ## WHEN checking the vcf
    ## THEN assert that the function raises a VcfError
    with pytest.raises(VcfError):
        check_vcf(unsorted_vcf_path)

def test_check_sv_vcf(sv_vcf_path):
    ## GIVEN a vcf file and a counter that checks the number of variants
    true_nr = 0
    
    with open(sv_vcf_path, 'r') as f:
        for line in f:
            if not line.startswith('#'):
                true_nr += 1
    
    ## WHEN collecting the VCF info
    vcf_info = check_vcf(sv_vcf_path, 'sv')
    
    ## THEN assert that the number of variants collected is correct
    assert vcf_info['nr_variants'] == true_nr
    ## THEN assert that the variant type is correct
    assert vcf_info['variant_type'] == 'sv'

def test_check_vcf_wrong_type(sv_vcf_path):
    ## GIVEN a sv vcf file
    
    ## WHEN collecting the VCF info with wrong variant type
    ## THEN assert that a VcfError is raised
    with pytest.raises(VcfError):
        vcf_info = check_vcf(sv_vcf_path, 'snv')



import logging
import pytest

import cyvcf2

from mongomock import MongoClient as MockClient
from pymongo import MongoClient
from ped_parser import FamilyParser

from loqusdb.plugins import MongoAdapter
from loqusdb.log import init_log
from loqusdb.models import Case
from loqusdb.build_models import (build_variant, build_case)
from loqusdb.utils.load import update_case

logger = logging.getLogger('.')

init_log(logger, loglevel='DEBUG')

REAL_DATABASE = 'test'

class CyvcfVariant(object):
    """Mock a cyvcf variant

    Default is to return a variant with three individuals high genotype
    quality.
    """
    def __init__(self, chrom='1', pos=80000, ref='A', alt='C', end=None,
                 gt_quals=[60, 60, 60], gt_types=[1, 1, 0], var_type='snv',
                 var_id=None, info_dict={}):
        super(CyvcfVariant, self).__init__()
        self.CHROM = chrom
        self.POS = pos
        self.REF = ref
        self.ALT = [alt]
        self.end = end or pos
        self.gt_quals = gt_quals
        self.gt_types = gt_types
        self.var_type = var_type
        self.INFO = info_dict
        if var_id is None:
            var_id = '.'
        self.ID = var_id

@pytest.fixture(scope='function')
def mongo_client(request):
    """Return a mongomock client"""
    client = MockClient()
    return client

@pytest.fixture(scope='function')
def real_mongo_client(request):
    """Return a mongomock client"""
    client = MongoClient()
    def teardown():
        print('\n')
        logger.info("Deleting database")
        client.drop_database(REAL_DATABASE)
        logger.info("Database deleted")

    request.addfinalizer(teardown)

    return client

@pytest.fixture(scope='function')
def mongo_adapter(request, mongo_client):
    """Return a mongo adapter"""
    db_name = 'test'
    adapter = MongoAdapter(mongo_client, db_name)

    return adapter

@pytest.fixture(scope='function')
def real_mongo_adapter(request, real_mongo_client):
    """Return a mongo adapter"""
    db_name = 'test'
    adapter = MongoAdapter(real_mongo_client, db_name)

    return adapter

@pytest.fixture(scope='function')
def simplest_variant(request):
    variant = {
        '_id': 'test',
    }
    return variant

@pytest.fixture(scope='function')
def homozygous_variant(request):
    variant = {
        '_id': 'test',
        'homozygote': 1,
        'case_id': '1'
    }
    return variant

@pytest.fixture(scope='function')
def simple_case(request):
    case = Case('test', './test.vcf')

    return case

@pytest.fixture(scope='function')
def vcf_path(request):
    file_path = 'tests/fixtures/test.vcf'
    return file_path

@pytest.fixture(scope='function')
def sv_vcf_path(request):
    file_path = 'tests/fixtures/test.SV.vcf'
    return file_path

@pytest.fixture(scope='function')
def double_vcf_path(request):
    "This will return a vcf with the same variant two times"
    file_path = 'tests/fixtures/double_variant.vcf'
    return file_path

@pytest.fixture(scope='function')
def unsorted_vcf_path(request):
    "This will return a vcf with unsorted variants"
    file_path = 'tests/fixtures/unsorted.vcf'
    return file_path

@pytest.fixture(scope='function')
def zipped_vcf_path(request):
    "Returns a VCF that is gzipped and have a index companion"
    file_path = 'tests/fixtures/test.vcf.gz'
    return file_path

@pytest.fixture(scope='function')
def ped_path(request):
    file_path = 'tests/fixtures/recessive_trio.ped'
    return file_path

@pytest.fixture(scope='function')
def funny_ped_path(request):
    "Returns a ped file with incorrect family relations"
    file_path = 'tests/fixtures/funny_trio.ped'
    return file_path

@pytest.fixture(scope='function')
def ind_positions(request):
    return {'proband': 0, 'mother':1, 'father':2}

@pytest.fixture(scope='function')
def case_lines(request, ped_path):
    """Return ped formated case lines"""
    case = []
    with open(ped_path, 'r') as f:
        for line in f:
            case.append(line)

    return case

@pytest.fixture(scope='function')
def vcf_obj(request, vcf_path):
    """return a cyvcf2.VCF obj"""
    return cyvcf2.VCF(vcf_path)

@pytest.fixture(scope='function')
def sv_vcf_obj(request, sv_vcf_path):
    """return a cyvcf2.VCF obj"""
    return cyvcf2.VCF(sv_vcf_path)

@pytest.fixture(scope='function')
def case_obj(request, case_lines, vcf_obj, vcf_path):
    """Return a case obj"""
    family_parser = FamilyParser(case_lines, family_type='ped')
    families = list(family_parser.families.keys())
    family = family_parser.families[families[0]]
    vcf_individuals = vcf_obj.samples
    nr_variants = 0
    for nr_variants,variant in enumerate(vcf_obj,1):
        pass
    _case_obj = build_case(
        case=family,
        vcf_individuals=vcf_individuals,
        vcf_path=vcf_path,
        nr_variants=nr_variants,
        )
    return _case_obj

@pytest.fixture(scope='function')
def sv_case_obj(request, case_lines, sv_vcf_obj, sv_vcf_path):
    """Return a case obj"""
    family_parser = FamilyParser(case_lines, family_type='ped')
    families = list(family_parser.families.keys())
    family = family_parser.families[families[0]]
    vcf_individuals = sv_vcf_obj.samples
    nr_variants = 0
    for nr_variants,variant in enumerate(sv_vcf_obj,1):
        pass
    _case_obj = build_case(
        case=family,
        sv_individuals=vcf_individuals,
        vcf_sv_path=sv_vcf_path,
        nr_sv_variants=nr_variants,
        )
    return _case_obj

@pytest.fixture(scope='function')
def complete_case_obj(request, case_obj, sv_case_obj):
    """Return a case obj with both sv and snv information"""
    _case_obj = update_case(case_obj, sv_case_obj)
    return _case_obj


@pytest.fixture(scope='function')
def family_obj(request, case_lines):
    """Return a case obj"""
    family_parser = FamilyParser(case_lines, family_type='ped')
    families = list(family_parser.families.keys())
    family = family_parser.families[families[0]]
    return family

@pytest.fixture(scope='function')
def case_id(request, case_lines):
    """Return a case id"""
    family_parser = FamilyParser(case_lines, family_type='ped')
    families = list(family_parser.families.keys())
    family = family_parser.families[families[0]]
    family_id = family.family_id

    return family_id

@pytest.fixture(scope='function')
def individuals(request, case_obj):
    """Return a case obj"""

    return case_obj.individuals

@pytest.fixture(scope='function')
def two_cases(request):
    """Return ped formated case lines"""
    case_lines = [
        "#FamilyID\tSampleID\tFather\tMother\tSex\tPhenotype\n",
        "1\tproband\t0\t0\t1\t2\n",
        "2\tproband\t0\t0\t2\t1\n",
    ]
    return case_lines

## Variant fixtures:

# chrom='1',
# pos=80000,
# ref='A',
# alt='C',
# end=None,
# gt_quals=[60, 60, 60],
# gt_types=[1, 1, 0],
# var_type='snv',
# info_dict={}

### Variant objects

@pytest.fixture(scope='function')
def variant_obj(request, het_variant, ind_positions, individuals):
    _variant_obj = build_variant(
        variant=het_variant,
        individuals = individuals,
        ind_positions = ind_positions,
        case_id='test',
        gq_treshold=None
    )
    return _variant_obj


### CYVCF2 variants
### SNVs:
@pytest.fixture(scope='function')
def hem_variant(request):
    variant_object = CyvcfVariant(chrom='X', pos=60000)
    return variant_object

@pytest.fixture(scope='function')
def variant_chr(request):
    variant_object = CyvcfVariant(chrom='chrX', pos=60000)
    return variant_object

@pytest.fixture(scope='function')
def par_variant(request):
    variant_object = CyvcfVariant(chrom='X', pos=60001)
    return variant_object

@pytest.fixture(scope='function')
def het_variant(request):
    variant_object = CyvcfVariant()
    return variant_object

@pytest.fixture(scope='function')
def variant_no_gq(request):
    variant_object = CyvcfVariant(gt_quals=[-1,-1,-1])
    return variant_object

@pytest.fixture(scope='function')
def hom_variant(request):
    variant_object = CyvcfVariant(gt_types=[3,1,1])
    return variant_object

@pytest.fixture(scope='function')
def variant_no_call(request):
    variant_object = CyvcfVariant(gt_types=[2, 2, 2])
    return variant_object

### SVs:
@pytest.fixture(scope='function')
def del_variant(request):
    variant_object = CyvcfVariant(
        chrom='1',
        ref='G',
        alt='<DEL>',
        pos=1285001,
        end=1287000,
        var_type='sv',
        info_dict={
            'END': 1287000,
            'SVLEN': -20000,
            'SVTYPE': 'DEL'
        }
    )
    return variant_object

@pytest.fixture(scope='function')
def small_insert_variant(request):
    variant_object = CyvcfVariant(
        chrom='1',
        ref='G',
        alt='GGGACGGGGGTTCTGAGATAAGCAAGCCCCCACCAGGTGAGACCGGCGGAGCTGTGGCCACCGAGGTCCCGGGAGCTGGTGCT',
        pos=3021145,
        end=3021145,
        var_type='sv',
        info_dict={
            'END': 3021145,
            'SVLEN': 82,
            'SVTYPE': 'INS'
        }
    )
    return variant_object

@pytest.fixture(scope='function')
def insertion_variant(request):
    variant_object = CyvcfVariant(
        chrom='1',
        ref='A',
        alt='<INS>',
        pos=3177306,
        end=3177306,
        var_type='sv',
        info_dict={
            'END': 3177306,
            'SVLEN': None,
            'SVTYPE': 'INS'
        }
    )
    return variant_object

@pytest.fixture(scope='function')
def duptandem_variant(request):
    variant_object = CyvcfVariant(
        chrom='1',
        ref='A',
        alt='<DUP:TANDEM>',
        pos=3092626,
        end=3092849,
        var_type='sv',
        info_dict={
            'END': 3092849,
            'SVLEN': 223,
            'SVTYPE': 'DUP'
        }
    )
    return variant_object

@pytest.fixture(scope='function')
def translocation_variant(request):
    variant_object = CyvcfVariant(
        chrom='1',
        ref='N',
        alt='N[11:119123896[',
        pos=3754913,
        end=3754913,
        var_type='sv',
        info_dict={
            'END': None,
            'SVLEN': None,
            'SVTYPE': 'BND'
        }
    )
    return variant_object




class Identity(dict):
    """Identity representation. This is used to ask benchmark queries"""
    def __init__(self, cluster_id, variant_id, case_id):
        """Construct a identity object

            Args:
                cluster_id(str): Ref to a cluster
                variant_id (str): ID from variant
                case_id (str): What case it belongs to
        """
        super(Identity, self).__init__(
            cluster_id=cluster_id,
            variant_id=variant_id,
            case_id=case_id,
        )



from pprint import pprint as pp

import pytest

from loqusdb.build_models.case import (build_case, get_individual_positions)
from loqusdb.exceptions import CaseError

def test_get_individual_positions():
    ## GIVEN a list with ids
    inds = ['1','2','3']
    ## WHEN getting the individual positions
    ind_pos = get_individual_positions(inds)
    ## THEN assert they where given the correct position
    assert ind_pos['1'] == 0
    assert ind_pos['2'] == 1
    assert ind_pos['3'] == 2

def test_get_individual_positions_no_inds():
    ## GIVEN a list with ids
    inds = None
    ## WHEN getting the individual positions
    ind_pos = get_individual_positions(inds)
    ## THEN assert an empty dict is returned
    assert ind_pos == {}
    
def test_build_case_no_ped():
    ## GIVEN some vcf individuals
    
    vcf_individuals = ['mother', 'proband']
    case_id = 'test'
    
    ## WHEN building a case object
    case_obj = build_case(
        case = None, 
        vcf_individuals=vcf_individuals, 
        case_id=case_id, 
    )
    
    ## THEN assert that the case got the right ID
    assert case_obj['case_id'] == case_id
    for ind_obj in case_obj['individuals']:
        assert ind_obj['name'] in vcf_individuals
        assert ind_obj['ind_id'] in vcf_individuals

def test_build_case_no_ped_no_case_id():
    ## GIVEN some vcf individuals
    
    vcf_individuals = ['mother', 'proband']
    
    ## WHEN building a case object

    ## THEN assert a CaseError is raised
    with pytest.raises(CaseError):
        case_obj = build_case(
            case = None, 
            vcf_individuals=vcf_individuals, 
        )

def test_build_case_ped(family_obj, vcf_path):
    ## GIVEN a ped parser family_obj
    vcf_inds = [ind_id for ind_id in family_obj.individuals]
    nr_variants = 10
    
    ## WHEN building a case object
    case_obj = build_case(
        case = family_obj, 
        vcf_individuals=vcf_inds,
        vcf_path = vcf_path,
        nr_variants = nr_variants,
    )
    
    ## THEN assert that the case has the correct id 
    assert case_obj['case_id'] == family_obj.family_id
    
    for ind_obj in case_obj['individuals']:
        assert ind_obj['ind_id'] in vcf_inds
    
    ## THEN assert that the vcf_path was added
    assert case_obj['vcf_path'] == vcf_path

    ## THEN assert that the nr variants is correct
    assert case_obj['nr_variants'] == nr_variants

def test_build_case_ped_sv(family_obj, sv_vcf_path):
    ## GIVEN a ped parser family_obj
    vcf_inds = [ind_id for ind_id in family_obj.individuals]
    nr_sv_variants = 10
    
    ## WHEN building a case object
    case_obj = build_case(
        case = family_obj, 
        sv_individuals=vcf_inds,
        vcf_sv_path = sv_vcf_path,
        nr_sv_variants = nr_sv_variants,
    )
    
    ## THEN assert that the case has the correct id 
    assert case_obj['case_id'] == family_obj.family_id
    
    case_obj['individuals'] == []
    for ind_obj in case_obj['sv_individuals']:
        assert ind_obj['ind_id'] in vcf_inds
    
    ## THEN assert that the vcf_path was added
    assert case_obj['vcf_path'] == None
    assert case_obj['vcf_sv_path'] == sv_vcf_path

    ## THEN assert that the nr variants is correct
    assert case_obj['nr_variants'] == None
    assert case_obj['nr_sv_variants'] == nr_sv_variants

def test_build_case_ped_sv_and_snv(family_obj, sv_vcf_path, vcf_path):
    ## GIVEN a ped parser family_obj
    vcf_inds = [ind_id for ind_id in family_obj.individuals]
    sv_vcf_inds = [ind_id for ind_id in family_obj.individuals]
    nr_sv_variants = 10
    nr_variants = 20
    
    ## WHEN building a case object
    case_obj = build_case(
        case = family_obj, 
        sv_individuals=vcf_inds,
        vcf_sv_path = sv_vcf_path,
        nr_sv_variants = nr_sv_variants,
        vcf_individuals=vcf_inds,
        vcf_path = vcf_path,
        nr_variants = nr_variants,
    )
    
    ## THEN assert that the case has the correct id 
    assert case_obj['case_id'] == family_obj.family_id
    
    for ind_obj in case_obj['individuals']:
        assert ind_obj['ind_id'] in vcf_inds

    for ind_obj in case_obj['sv_individuals']:
        assert ind_obj['ind_id'] in sv_vcf_inds
    
    ## THEN assert that the vcf_path was added
    assert case_obj['vcf_path'] == vcf_path
    assert case_obj['vcf_sv_path'] == sv_vcf_path

    ## THEN assert that the nr variants is correct
    assert case_obj['nr_variants'] == nr_variants
    assert case_obj['nr_sv_variants'] == nr_sv_variants



import logging

from pprint import pprint as pp

from loqusdb.plugins import BaseVariantMixin
from .structural_variant import SVMixin

from pymongo import (ASCENDING, DESCENDING)

LOG = logging.getLogger(__name__)

class VariantMixin(BaseVariantMixin, SVMixin):
    
    def add_variant(self, variant):
        """Add a variant to the variant collection
        
            If the variant exists we update the count else we insert a new variant object.
        
            Args:
                variant (dict): A variant dictionary
        
        """
        LOG.debug("Upserting variant: {0}".format(variant.get('_id')))
        
        
        update = {
                '$inc': {
                    'homozygote': variant.get('homozygote', 0),
                    'hemizygote': variant.get('hemizygote', 0),
                    'observations': 1
                },
                '$set': {
                    'chrom': variant.get('chrom'),
                    'start': variant.get('pos'),
                    'end': variant.get('end'),
                    'ref': variant.get('ref'),
                    'alt': variant.get('alt'),
                }
             }
        if variant.get('case_id'):
            update['$push'] = {
                                'families': {
                                '$each': [variant.get('case_id')],
                                '$slice': -50
                                }
                            }

        message = self.db.variant.update_one(
            {'_id': variant['_id'],},
            update,
             upsert=True
        )
        if message.modified_count == 1:
            LOG.debug("Variant %s was updated", variant.get('_id'))
        else:
            LOG.debug("Variant was added to database for first time")
        return

    def get_variant(self, variant):
        """Check if a variant exists in the database and return it.
        Variants are searched with a variant id which is a string that consists of
        chrom_pos_ref_alt. 
        There is no simple way to create a similar ID for structural variants so this function
        will not work for SVs

        Search the variants with the variant id

        Args:
            variant(dict): A variant dictionary

        Returns:
            variant(dict): A variant dictionary or None
        """
        return self.db.variant.find_one({'_id': variant.get('_id')})

    def get_variants(self, chromosome=None, start=None, end=None):
        """Return all variants in the database
        If no region is specified all variants will be returned.

        Args:
            chromosome(str)
            start(int)
            end(int)
        
    
        Returns:
            variants(Iterable(Variant))
        """
        query = {}
        if chromosome:
            query['chrom'] = chromosome
        if start:
            query['start'] = {'$lte': end}
            query['end'] = {'$gte': start}
        LOG.info("Find all variants {}".format(query))
        return self.db.variant.find(query).sort([('start', ASCENDING)])

    def delete_variant(self, variant):
        """Delete observation in database

        This means that we take down the observations variable with one.
        If 'observations' == 1 we remove the variant. If variant was homozygote
        we decrease 'homozygote' with one.
        Also remove the family from array 'families'.

        Args:
            variant (dict): A variant dictionary

        """
        mongo_variant = self.get_variant(variant)
        
        if mongo_variant:
            
            if mongo_variant['observations'] == 1:
                LOG.debug("Removing variant {0}".format(
                    mongo_variant.get('_id')
                ))
                message = self.db.variant.delete_one({'_id': variant['_id']})
            else:
                LOG.debug("Decreasing observations for {0}".format(
                    mongo_variant.get('_id')
                ))
                message = self.db.variant.update_one({
                    '_id': mongo_variant['_id']
                    },{
                        '$inc': {
                            'observations': -1,
                            'homozygote': - (variant.get('homozygote', 0)),
                            'hemizygote': - (variant.get('hemizygote', 0)),
                        },
                        '$pull': {
                            'families': variant.get('case_id')
                        }
                    }, upsert=False)
        return


    def get_chromosomes(self, sv=False):
        """Return a list of all chromosomes found in database
        
        Args:
            sv(bool): if sv variants should be choosen
        
        Returns:
            res(iterable(str)): An iterable with all chromosomes in the database
        """
        if sv:
            res = self.db.structural_variant.distinct('chrom')
        else:
            res = self.db.variant.distinct('chrom')
            
        return res
    
    def get_max_position(self, chrom):
        """Get the last position observed on a chromosome in the database
        
        Args:
            chrom(str)
        
        Returns:
            end(int): The largest end position found
            
        """
        res = self.db.variant.find({'chrom':chrom}, {'_id':0, 'end':1}).sort([('end', DESCENDING)]).limit(1)
        end = 0
        for variant in res:
            end = variant['end']
        return end
    


import logging

from loqusdb.models import (Case, Individual)
from loqusdb.exceptions import CaseError

LOG = logging.getLogger(__name__)

def get_individual_positions(individuals):
    """Return a dictionary with individual positions

    Args:
        individuals(list): A list with vcf individuals in correct order

    Returns:
        ind_pos(dict): Map from ind_id -> index position
    """
    ind_pos = {}
    if individuals:
        for i, ind in enumerate(individuals):
            ind_pos[ind] = i
    return ind_pos

def build_case(case, vcf_individuals=None, case_id=None, vcf_path=None, sv_individuals=None,
               vcf_sv_path=None, nr_variants=None, nr_sv_variants=None):
    """Build a Case from the given information

    Args:
        case(ped_parser.Family): A family object
        vcf_individuals(list): Show the order of inds in vcf file
        case_id(str): If another name than the one in family file should be used
        vcf_path(str)
        sv_individuals(list): Show the order of inds in vcf file
        vcf_sv_path(str)
        nr_variants(int)
        nr_sv_variants(int)

    Returns:
        case_obj(models.Case)
    """
    # Create a dict that maps the ind ids to the position they have in vcf
    individual_positions = get_individual_positions(vcf_individuals)
    sv_individual_positions = get_individual_positions(sv_individuals)

    family_id = None
    if case:
        if not case.affected_individuals:
            LOG.warning("No affected individuals could be found in ped file")
        family_id = case.family_id

    # If case id is given manually we use that one
    case_id = case_id or family_id
    if case_id is None:
        raise CaseError

    case_obj = Case(
        case_id=case_id,
    )

    if vcf_path:
        case_obj['vcf_path'] = vcf_path
        case_obj['nr_variants'] = nr_variants

    if vcf_sv_path:
        case_obj['vcf_sv_path'] = vcf_sv_path
        case_obj['nr_sv_variants'] = nr_sv_variants

    ind_objs = []
    if case:
        if individual_positions:
            _ind_pos = individual_positions
        else:
            _ind_pos = sv_individual_positions

        for ind_id in case.individuals:
            individual = case.individuals[ind_id]
            try:
                ind_obj = Individual(
                    ind_id=ind_id,
                    case_id=case_id,
                    ind_index=_ind_pos[ind_id],
                    sex=individual.sex,
                )
                ind_objs.append(dict(ind_obj))
            except KeyError:
                raise CaseError("Ind %s in ped file does not exist in VCF", ind_id)
    else:
        # If there where no family file we can create individuals from what we know
        for ind_id in individual_positions:
            ind_obj = Individual(
                ind_id = ind_id,
                case_id = case_id,
                ind_index=individual_positions[ind_id],
            )
            ind_objs.append(dict(ind_obj))

    # Add individuals to the correct variant type
    for ind_obj in ind_objs:
        if vcf_sv_path:
            case_obj['sv_individuals'].append(dict(ind_obj))
            case_obj['_sv_inds'][ind_obj['ind_id']] = dict(ind_obj)
        if vcf_path:
            case_obj['individuals'].append(dict(ind_obj))
            case_obj['_inds'][ind_obj['ind_id']] = dict(ind_obj)

    return case_obj



from loqusdb.build_models.variant import build_variant

def test_get_insertion(small_insert_variant, mongo_adapter, case_obj):
    adapter = mongo_adapter
    ## GIVEN a mongo adapter with a small insertion
    variant = small_insert_variant
    case_id = case_obj['case_id']
    formated_variant = build_variant(
        variant=variant,
        case_obj=case_obj,
        case_id=case_id
    )
    
    adapter.add_case(case_obj)
    adapter.add_structural_variant(formated_variant)
    for variant_obj in adapter.db.structural_variant.find():
        assert variant_obj

def test_get_translocation(translocation_variant, mongo_adapter, case_obj):
    adapter = mongo_adapter
    ## GIVEN a mongo adapter with a translocation
    variant = translocation_variant
    case_id = case_obj['case_id']
    formated_variant = build_variant(
        variant=variant,
        case_obj=case_obj,
        case_id=case_id
    )
    
    adapter.add_case(case_obj)
    adapter.add_structural_variant(formated_variant)
    for variant_obj in adapter.db.structural_variant.find():
        assert variant_obj



# -*- coding: utf-8 -*-
"""
loqusdb.utils.update.py

Functions to update data into the database.
This functions take an adapter which is the communication device for the database.

"""

import logging

from pprint import pprint as pp

from .vcf import (get_vcf, check_vcf)
from .case import (get_case, update_case)
from .delete import delete
from .load import (load_case,load_variants)
from loqusdb.build_models import (build_case, build_variant)
from loqusdb.exceptions import (CaseError, VcfError)

LOG = logging.getLogger(__name__)

def update_database(adapter, variant_file=None, sv_file=None, family_file=None, family_type='ped', 
           skip_case_id=False, gq_treshold=None, case_id=None, max_window = 3000):
    """Update a case in the database
            
    Args:
          adapter: Connection to database
          variant_file(str): Path to variant file
          sv_file(str): Path to sv variant file
          family_file(str): Path to family file
          family_type(str): Format of family file
          skip_case_id(bool): If no case information should be added to variants
          gq_treshold(int): If only quality variants should be considered
          case_id(str): If different case id than the one in family file should be used
          max_window(int): Specify the max size for sv windows

    Returns:
          nr_inserted(int)
    """
    vcf_files = []
    nr_variants = None
    vcf_individuals = None
    if variant_file:
        vcf_info = check_vcf(variant_file)
        nr_variants = vcf_info['nr_variants']
        variant_type = vcf_info['variant_type']
        vcf_files.append(variant_file)
        # Get the indivuduals that are present in vcf file
        vcf_individuals = vcf_info['individuals']

    nr_sv_variants = None
    sv_individuals = None
    if sv_file:
        vcf_info = check_vcf(sv_file, 'sv')
        nr_sv_variants = vcf_info['nr_variants']
        vcf_files.append(sv_file)
        sv_individuals = vcf_info['individuals']

    # If a gq treshold is used the variants needs to have GQ
    for _vcf_file in vcf_files:
        # Get a cyvcf2.VCF object
        vcf = get_vcf(_vcf_file)
        
        if gq_treshold:
            if not vcf.contains('GQ'):
                LOG.warning('Set gq-treshold to 0 or add info to vcf {0}'.format(_vcf_file))
                raise SyntaxError('GQ is not defined in vcf header')

    # Get a ped_parser.Family object from family file
    family = None
    family_id = None
    if family_file:
        with open(family_file, 'r') as family_lines:
            family = get_case(
                family_lines=family_lines, 
                family_type=family_type
            )
            family_id = family.family_id
    
    # There has to be a case_id or a family at this stage.
    case_id = case_id or family_id

    # Convert infromation to a loqusdb Case object
    case_obj = build_case(
        case=family, 
        case_id=case_id,
        vcf_path=variant_file,
        vcf_individuals=vcf_individuals,
        nr_variants=nr_variants,
        vcf_sv_path=sv_file,
        sv_individuals=sv_individuals,
        nr_sv_variants=nr_sv_variants,
    )

    existing_case = adapter.case(case_obj)
    if not existing_case:
        raise CaseError("Case {} does not exist in database".format(case_obj['case_id']))

    # Update the existing case in database
    case_obj = load_case(
        adapter=adapter,
        case_obj=case_obj,
        update=True,
        )
    
    nr_inserted = 0
    # If case was succesfully added we can store the variants
    for file_type in ['vcf_path','vcf_sv_path']:
        variant_type = 'snv'
        if file_type == 'vcf_sv_path':
            variant_type = 'sv'
        if case_obj.get(file_type) is None:
            continue

        vcf_obj = get_vcf(case_obj[file_type])
        try:
            nr_inserted += load_variants(  
                adapter=adapter,
                vcf_obj=vcf_obj,
                case_obj=case_obj, 
                skip_case_id=skip_case_id,
                gq_treshold=gq_treshold,
                max_window=max_window,
                variant_type=variant_type,
            )
        except Exception as err:
            # If something went wrong do a rollback
            LOG.warning(err)
            delete(
                adapter=adapter,
                case_obj=case_obj,
                update=True,
                existing_case=existing_case,
            )
            raise err
    return nr_inserted



# -*- coding: utf-8 -*-
"""
loqusdb.utils.load.py

Functions to load data into the database.
This functions take an adapter which is the communication device for the database.

"""

import logging
import click
from datetime import datetime

from pprint import pprint as pp

from .vcf import (get_vcf, check_vcf)
from .case import (get_case, update_case)
from .delete import delete
from loqusdb.build_models import (build_case, build_variant)
from loqusdb.exceptions import (CaseError, VcfError)

LOG = logging.getLogger(__name__)

def load_database(adapter, variant_file=None, sv_file=None, family_file=None,
                  family_type='ped', skip_case_id=False, gq_treshold=None,
                  case_id=None, max_window = 3000):
    """Load the database with a case and its variants

    Args:
          adapter: Connection to database
          variant_file(str): Path to variant file
          sv_file(str): Path to sv variant file
          family_file(str): Path to family file
          family_type(str): Format of family file
          skip_case_id(bool): If no case information should be added to variants
          gq_treshold(int): If only quality variants should be considered
          case_id(str): If different case id than the one in family file should be used
          max_window(int): Specify the max size for sv windows

    Returns:
          nr_inserted(int)
    """
    vcf_files = []

    nr_variants = None
    vcf_individuals = None
    if variant_file:
        vcf_info = check_vcf(variant_file)
        nr_variants = vcf_info['nr_variants']
        variant_type = vcf_info['variant_type']
        vcf_files.append(variant_file)
        # Get the indivuduals that are present in vcf file
        vcf_individuals = vcf_info['individuals']

    nr_sv_variants = None
    sv_individuals = None
    if sv_file:
        vcf_info = check_vcf(sv_file, 'sv')
        nr_sv_variants = vcf_info['nr_variants']
        vcf_files.append(sv_file)
        sv_individuals = vcf_info['individuals']

    # If a gq treshold is used the variants needs to have GQ
    for _vcf_file in vcf_files:
        # Get a cyvcf2.VCF object
        vcf = get_vcf(_vcf_file)

        if gq_treshold:
            if not vcf.contains('GQ'):
                LOG.warning('Set gq-treshold to 0 or add info to vcf {0}'.format(_vcf_file))
                raise SyntaxError('GQ is not defined in vcf header')

    # Get a ped_parser.Family object from family file
    family = None
    family_id = None
    if family_file:
        LOG.info("Loading family from %s", family_file)
        with open(family_file, 'r') as family_lines:
            family = get_case(
                family_lines=family_lines,
                family_type=family_type
            )
            family_id = family.family_id

    # There has to be a case_id or a family at this stage.
    case_id = case_id or family_id
    # Convert infromation to a loqusdb Case object
    case_obj = build_case(
        case=family,
        case_id=case_id,
        vcf_path=variant_file,
        vcf_individuals=vcf_individuals,
        nr_variants=nr_variants,
        vcf_sv_path=sv_file,
        sv_individuals=sv_individuals,
        nr_sv_variants=nr_sv_variants,
    )
    # Build and load a new case, or update an existing one
    load_case(
        adapter=adapter,
        case_obj=case_obj,
    )

    nr_inserted = 0
    # If case was succesfully added we can store the variants
    for file_type in ['vcf_path','vcf_sv_path']:
        variant_type = 'snv'
        if file_type == 'vcf_sv_path':
            variant_type = 'sv'
        if case_obj.get(file_type) is None:
            continue

        vcf_obj = get_vcf(case_obj[file_type])
        try:
            nr_inserted += load_variants(
                adapter=adapter,
                vcf_obj=vcf_obj,
                case_obj=case_obj,
                skip_case_id=skip_case_id,
                gq_treshold=gq_treshold,
                max_window=max_window,
                variant_type=variant_type,
            )
        except Exception as err:
            # If something went wrong do a rollback
            LOG.warning(err)
            delete(
                adapter=adapter,
                case_obj=case_obj,
            )
            raise err
    return nr_inserted

def load_case(adapter, case_obj, update=False):
    """Load a case to the database

    Args:
        adapter: Connection to database
        case_obj: dict
        update(bool): If existing case should be updated

    Returns:
        case_obj(models.Case)
    """
    # Check if the case already exists in database.
    existing_case = adapter.case(case_obj)
    if existing_case:
        if not update:
            raise CaseError("Case {0} already exists in database".format(case_obj['case_id']))
        case_obj = update_case(case_obj, existing_case)

    # Add the case to database
    try:
        adapter.add_case(case_obj, update=update)
    except CaseError as err:
        raise err

    return case_obj

def load_variants(adapter, vcf_obj, case_obj, skip_case_id=False, gq_treshold=None,
                  max_window=3000, variant_type='snv'):
    """Load variants for a family into the database.

    Args:
        adapter (loqusdb.plugins.Adapter): initialized plugin
        case_obj(Case): dict with case information
        nr_variants(int)
        skip_case_id (bool): whether to include the case id on variant level
                             or not
        gq_treshold(int)
        max_window(int): Specify the max size for sv windows
        variant_type(str): 'sv' or 'snv'

    Returns:
        nr_inserted(int)
    """
    if variant_type == 'snv':
        nr_variants = case_obj['nr_variants']
    else:
        nr_variants = case_obj['nr_sv_variants']

    nr_inserted = 0
    case_id = case_obj['case_id']
    if skip_case_id:
        case_id = None
    # Loop over the variants in the vcf
    with click.progressbar(vcf_obj, label="Inserting variants",length=nr_variants) as bar:
        for variant in bar:
            #Creates a variant that is ready to insert into the database
            formated_variant = build_variant(
                    variant=variant,
                    case_obj=case_obj,
                    case_id=case_id,
                    gq_treshold=gq_treshold,
                )
            # We need to check if there was any information returned
            # The variant could be excluded based on low gq or if no individiual was called
            # in the particular case
            if not formated_variant:
                continue
            if formated_variant['is_sv']:
                adapter.add_structural_variant(variant=formated_variant, max_window=max_window)
            else:
                adapter.add_variant(variant=formated_variant)
            nr_inserted += 1

    LOG.info("Inserted %s variants of type %s", nr_inserted, variant_type)

    return nr_inserted



import pytest
from loqusdb.models import Case
from loqusdb.exceptions import CaseError


def test_delete_family(mongo_adapter):
    ## GIVEN a mongo adapter with a case
    db = mongo_adapter.db
    case = {
        'case_id': '1',
        'vcf_path': 'path_to_vcf'
    }
    
    db.case.insert_one(case)
    mongo_case = db.case.find_one()
    assert mongo_case['case_id'] == case['case_id']
    
    ## WHEN deleting the case
    mongo_adapter.delete_case(case)
    
    ## THEN assert that the case was deleted
    mongo_case = db.case.find_one()
    
    assert mongo_case == None

def test_delete_non_existing_family(mongo_adapter):
    ## GIVEN a mongo adapter and a empty database
    case = {
        'case_id': '1',
        'vcf_path': 'path_to_vcf'
    }
    
    ## WHEN deleting a non existing case
    with pytest.raises(CaseError):
    ## THEN assert a CaseError is raised
        mongo_adapter.delete_case(case)



import os
import logging
import click

from pprint import pprint as pp

from datetime import datetime

from loqusdb.exceptions import (CaseError, VcfError)
from loqusdb.utils.update import update_database
from loqusdb.utils.vcf import (get_file_handle, check_vcf)

from . import base_command

LOG = logging.getLogger(__name__)

@base_command.command('update', short_help="Update an existing case with a new type of variants")
@click.option('--variant-file',
                    type=click.Path(exists=True),
                    metavar='<vcf_file>',
                    help="Load a VCF with SNV/INDEL Variants",
)
@click.option('--sv-variants',
                    type=click.Path(exists=True),
                    metavar='<sv_vcf_file>',
                    help="Load a VCF with Structural Variants",
)
@click.option('-f', '--family-file',
                    type=click.Path(exists=True),
                    metavar='<ped_file>'
)
@click.option('-t' ,'--family-type', 
                type=click.Choice(['ped', 'alt', 'cmms', 'mip']), 
                default='ped',
                show_default=True,
                help='If the analysis use one of the known setups, please specify which one.'
)
@click.option('-c' ,'--case-id', 
                type=str, 
                help='If a different case id than the one in ped file should be used'
)
@click.option('-s' ,'--skip-case-id', 
                is_flag=True, 
                show_default=True,
                help='Do not store case information on variants'
)
@click.option('--ensure-index', 
                is_flag=True, 
                help='Make sure that the indexes are in place'
)
@click.option('--gq-treshold',
    default=20,
    show_default=True,
    help='Treshold to consider variant'
)
@click.option('--max-window', '-m',
    default=2000,
    show_default=True,
    help='Specify the maximum window size for svs'
)
@click.pass_context
def update(ctx, variant_file, sv_variants, family_file, family_type, skip_case_id, gq_treshold, 
           case_id, ensure_index, max_window):
    """Load the variants of a case

    A variant is loaded if it is observed in any individual of a case
    If no family file is provided all individuals in vcf file will be considered.
    """
    if not (family_file or case_id):
        LOG.warning("Please provide a family file or a case id")
        ctx.abort()
    
    if not (variant_file or sv_variants):
        LOG.warning("Please provide a VCF file")
        ctx.abort()
    

    variant_path = None
    if variant_file:
        variant_path = os.path.abspath(variant_file)
        
    variant_sv_path = None
    if sv_variants:
        variant_sv_path = os.path.abspath(sv_variants)
    
    adapter = ctx.obj['adapter']

    start_inserting = datetime.now()
    
    
    try:
        nr_inserted = update_database(
            adapter=adapter,
            variant_file=variant_path,
            sv_file=variant_sv_path,
            family_file=family_file,
            family_type=family_type,
            skip_case_id=skip_case_id,
            case_id=case_id,
            gq_treshold=gq_treshold,
            max_window=max_window,
        )
    except (SyntaxError, CaseError, IOError, VcfError) as error:
        LOG.warning(error)
        ctx.abort()

    LOG.info("Nr variants inserted: %s", nr_inserted)
    LOG.info("Time to insert variants: {0}".format(
                datetime.now() - start_inserting))
    
    if ensure_index:
        adapter.ensure_indexes()
    else:
        adapter.check_indexes()
    



import pytest
from cyvcf2 import VCF

from loqusdb.utils.load import load_variants

def test_load_variants(real_mongo_adapter, het_variant, case_obj):
    mongo_adapter = real_mongo_adapter
    ## GIVEN an adapter and a vcf with one heterozygote variant
    db = mongo_adapter.db

    vcf = []
    vcf.append(het_variant)
    mongo_variant = db.variant.find_one()

    assert mongo_variant == None

    ## WHEN loading the variant into the database
    load_variants(
        adapter=mongo_adapter,
        vcf_obj=vcf,
        case_obj=case_obj,
    )

    mongo_variant = db.variant.find_one()

    ## THEN assert that the variant is loaded correct
    assert mongo_variant['families'] == [case_obj['case_id']]
    assert mongo_variant['observations'] == 1
    assert mongo_variant['homozygote'] == 0
    assert mongo_variant['hemizygote'] == 0

def test_load_homozygote(real_mongo_adapter, hom_variant, case_obj):
    mongo_adapter = real_mongo_adapter
    ## GIVEN an adapter and a vcf with one homozygote variant
    db = mongo_adapter.db

    vcf = []
    vcf.append(hom_variant)
    assert db.variant.find_one() == None

    ## WHEN loading the variant into the database
    load_variants(
        adapter=mongo_adapter,
        vcf_obj=vcf,
        case_obj=case_obj,
    )
    mongo_variant = db.variant.find_one()

    ## THEN assert that the variant is loaded correct
    assert mongo_variant['families'] == [case_obj['case_id']]
    assert mongo_variant['observations'] == 1
    assert mongo_variant['homozygote'] == 1
    assert mongo_variant['hemizygote'] == 0

def test_load_hemizygote(real_mongo_adapter, hem_variant, case_obj):
    mongo_adapter = real_mongo_adapter
    ## GIVEN an adapter and a vcf with one hemizygote variant
    db = mongo_adapter.db

    vcf = []
    vcf.append(hem_variant)

    assert db.variant.find_one() == None

    ## WHEN loading the variant into the database
    load_variants(
        adapter=mongo_adapter,
        vcf_obj=vcf,
        case_obj=case_obj,
    )
    mongo_variant = db.variant.find_one()

    ## THEN assert that the variant is loaded correct
    assert mongo_variant['families'] == [case_obj['case_id']]
    assert mongo_variant['observations'] == 1
    assert mongo_variant['homozygote'] == 0
    assert mongo_variant['hemizygote'] == 1

def test_load_par_variant(real_mongo_adapter, par_variant, case_obj):
    mongo_adapter = real_mongo_adapter

    ## GIVEN an adapter and a vcf with one PAR variant
    db = mongo_adapter.db

    vcf = []
    vcf.append(par_variant)

    assert db.variant.find_one() == None

    ## WHEN loading the variant into the database
    load_variants(
        adapter=mongo_adapter,
        vcf_obj=vcf,
        case_obj=case_obj,
    )
    mongo_variant = db.variant.find_one()

    ## THEN assert that the variant is loaded correct
    assert mongo_variant['families'] == [case_obj['case_id']]
    assert mongo_variant['observations'] == 1
    assert mongo_variant['homozygote'] == 0
    assert mongo_variant['hemizygote'] == 0

def test_load_two_variants(real_mongo_adapter, het_variant, case_obj):
    mongo_adapter = real_mongo_adapter
    ## GIVEN an adapter and a vcf with tho heterygote variants
    db = mongo_adapter.db

    vcf = []
    vcf.append(het_variant)
    vcf.append(het_variant)

    ## WHEN loading the variants into the database
    load_variants(
        adapter=mongo_adapter,
        vcf_obj=vcf,
        case_obj=case_obj,
    )

    ## THEN assert that the variant is loaded correct
    mongo_variant = db.variant.find_one()

    assert mongo_variant['observations'] == 2

def test_load_variants_skip_case_id(real_mongo_adapter, het_variant, case_obj):
    mongo_adapter = real_mongo_adapter
    ## GIVEN an adapter and a vcf with tho heterygote variants
    db = mongo_adapter.db

    vcf = []
    vcf.append(het_variant)

    ## WHEN loading the variants into the database
    load_variants(
        adapter=mongo_adapter,
        vcf_obj=vcf,
        case_obj=case_obj,
        skip_case_id=True,
    )

    mongo_variant = db.variant.find_one()

    ## THEN assert that the variant is loaded correct
    assert mongo_variant.get('families') == None

def test_load_same_variant_different_case(real_mongo_adapter, het_variant, case_obj):
    mongo_adapter = real_mongo_adapter
    ## GIVEN an adapter and a vcf
    db = mongo_adapter.db

    vcf = []
    vcf.append(het_variant)

    ## WHEN loading the variant into the database
    load_variants(
        adapter=mongo_adapter,
        vcf_obj=vcf,
        case_obj=case_obj,
    )

    case_id = case_obj['case_id']
    case_id2 = '2'
    case_obj['case_id'] = case_id2

    load_variants(
        adapter=mongo_adapter,
        vcf_obj=vcf,
        case_obj=case_obj,
    )

    mongo_variant = db.variant.find_one()

    assert mongo_variant['observations'] == 2
    assert mongo_variant['families'] == [case_id, case_id2]

def test_load_case_variants(real_mongo_adapter, case_obj):
    mongo_adapter = real_mongo_adapter

    db = mongo_adapter.db
    ## GIVEN a mongo adatper with snv variant file
    vcf_obj = VCF(case_obj['vcf_path'])
    ## WHEN loading the variants
    nr_variants = load_variants(
        adapter=mongo_adapter,
        vcf_obj=vcf_obj,
        case_obj=case_obj,
    )

    nr_loaded = 0
    for nr_loaded, variant in enumerate(db.variant.find(),1):
        pass
    ## THEN assert that the correct number of variants was loaded
    assert nr_loaded > 0
    assert nr_loaded == case_obj['nr_variants']

def test_load_sv_case_variants(mongo_adapter, sv_case_obj):
    db = mongo_adapter.db
    ## GIVEN a mongo adatper with snv variant file
    vcf_obj = VCF(sv_case_obj['vcf_sv_path'])
    ## WHEN loading the variants
    nr_variants = load_variants(
        adapter=mongo_adapter,
        vcf_obj=vcf_obj,
        case_obj=sv_case_obj,
        variant_type='sv',
    )

    nr_loaded_svs = 0
    for nr_loaded_svs, variant in enumerate(db.structural_variant.find(),1):
        pass
    nr_loaded_snvs = 0
    for nr_loaded_snvs, variant in enumerate(db.variant.find(),1):
        pass
    ## THEN assert that the correct number of variants was loaded
    assert nr_loaded_svs > 0
    assert nr_loaded_snvs == 0
    assert nr_loaded_svs == sv_case_obj['nr_sv_variants']



# -*- coding: utf-8 -*-
import logging
from datetime import datetime

from .vcf import get_vcf
from .case import get_case
from loqusdb.build_models import (build_case, build_variant)

LOG = logging.getLogger(__name__)


def delete(adapter, case_obj, update=False, existing_case=False):
    """Delete a case and all of it's variants from the database.
    
    Args:
        adapter: Connection to database
        case_obj(models.Case)
        update(bool): If we are in the middle of an update
        existing_case(models.Case): If something failed during an update we need to revert
                                    to the original case
    
    """
    # This will overwrite the updated case with the previous one
    if update:
        adapter.add_case(existing_case)
    else:
        adapter.delete_case(case_obj)

    for file_type in ['vcf_path','vcf_sv_path']:
        if not case_obj.get(file_type):
            continue
        variant_file = case_obj[file_type]
        # Get a cyvcf2.VCF object
        vcf_obj = get_vcf(variant_file)

        delete_variants(
            adapter=adapter,
            vcf_obj=vcf_obj,
            case_obj=case_obj,
        )

def delete_variants(adapter, vcf_obj, case_obj, case_id=None):
    """Delete variants for a case in the database
    
    Args:
        adapter(loqusdb.plugins.Adapter)
        vcf_obj(iterable(dict))
        ind_positions(dict)
        case_id(str)
    
    Returns:
        nr_deleted (int): Number of deleted variants
    """
    case_id = case_id or case_obj['case_id']
    nr_deleted = 0
    start_deleting = datetime.now()
    chrom_time = datetime.now()
    current_chrom = None
    new_chrom = None
    
    for variant in vcf_obj:
        formated_variant = build_variant(
            variant=variant,
            case_obj=case_obj,
            case_id=case_id,
        )
        
        if not formated_variant:
            continue
        
        new_chrom = formated_variant.get('chrom')
        adapter.delete_variant(formated_variant)
        nr_deleted += 1
        
        if not current_chrom:
            LOG.info("Start deleting chromosome {}".format(new_chrom))
            current_chrom = new_chrom
            chrom_time = datetime.now()
            continue
        
        if new_chrom != current_chrom:
            LOG.info("Chromosome {0} done".format(current_chrom))
            LOG.info("Time to delete chromosome {0}: {1}".format(
                current_chrom, datetime.now()-chrom_time))
            LOG.info("Start deleting chromosome {0}".format(new_chrom))
            current_chrom = new_chrom


    return nr_deleted



# -*- coding: utf-8 -*-
import logging
import click

from pprint import pprint as pp

from . import base_command

LOG = logging.getLogger(__name__)

@base_command.command('cases', short_help="Display cases in database")
@click.option('-c' ,'--case-id', 
                help='Search for case'
)
@click.option('--to-json', is_flag=True)
@click.pass_context
def cases(ctx, case_id, to_json):
    """Display cases in the database."""
    
    adapter = ctx.obj['adapter']
    cases = []
    
    if case_id:
        case_obj = adapter.case(case_id)
        if not case_obj:
            LOG.info("Case {0} does not exist in database".format(case_id))
            return
        cases.append(case_obj)
    else:
        cases = adapter.cases()
        if cases.count() == 0:
            LOG.info("No cases found in database")
            return
    
    if not to_json:
        click.echo("#case_id\tvcf_path")

    for case_obj in cases:
        if to_json:
            click.echo(case_obj)
        else:
            click.echo("{0}\t{1}".format(case_obj.get('case_id'), case_obj.get('vcf_path')))

@base_command.command('variants', short_help="Display variants in database")
@click.option('--variant-id', 
                help='Search for a variant'
)
@click.option('-c', '--chromosome', 
                help='Search for all variants in a chromosome'
)
@click.option('--end-chromosome', 
                help='Search for all variants that ends on chromosome'
)
@click.option('-s', '--start',
                help='Start of region',
                type=int
)
@click.option('-e', '--end', 
                help='End of region',
                type=int
)
@click.option('-t', '--variant-type', 
                help='Variant type to search for',
                type=click.Choice(['sv', 'snv']),
                default='snv'
)
@click.option('--sv-type', 
                help='Type of svs to search for',
)
# @click.option('--sort-key',
#                 help='Specify what field to sort on',
# )
@click.pass_context
def variants(ctx, variant_id, chromosome, end_chromosome, start, end, variant_type, 
             sv_type):
    """Display variants in the database."""
    if sv_type:
        variant_type = 'sv'
    
    adapter = ctx.obj['adapter']
    
    if (start or end):
        if not (chromosome and start and end):
            LOG.warning("Regions must be specified with chromosome, start and end")
            ctx.abort()
    
    if variant_id:
        variant = adapter.get_variant({'_id':variant_id})
        if variant:
            click.echo(variant)
        else:
            LOG.info("Variant {0} does not exist in database".format(variant_id))
        return
    
    if variant_type == 'snv':
        result = adapter.get_variants(
            chromosome=chromosome, 
            start=start, 
            end=end
        )
    else:
        LOG.info("Search for svs")
        result = adapter.get_sv_variants(
            chromosome=chromosome, 
            end_chromosome=end_chromosome, 
            sv_type=sv_type, 
            pos=start, 
            end=end
        )
        
    i = 0
    for variant in result:
        i += 1
        pp(variant)
    
    LOG.info("Number of variants found in database: %s", i)

@base_command.command('index', short_help="Add indexes to database")
@click.option('--view', 
    is_flag=True,
    help='Only display existing indexes',
)
@click.pass_context
def index(ctx, view):
    """Index the database."""
    adapter = ctx.obj['adapter']
    if view:
        click.echo(adapter.indexes())
        return
    adapter.ensure_indexes()



from loqusdb.build_models.variant import (build_variant, GENOTYPE_MAP)
import pytest


def test_format_variant(het_variant, case_obj):
    ## GIVEN a parsed variant
    variant = het_variant
    case_id = case_obj['case_id']
    ## WHEN parsing the variant
    formated_variant = build_variant(
        variant=variant,
        case_obj=case_obj,
        case_id=case_id
        )
    
    expected_id = '_'.join([
        variant.CHROM,
        str(variant.POS), 
        variant.REF, 
        variant.ALT[0]
    ])
    
    ## THEN assert it was built in a correct way
    assert formated_variant
    assert formated_variant['variant_id'] == expected_id
    assert formated_variant['chrom'] == variant.CHROM
    assert formated_variant['pos'] == variant.POS
    assert formated_variant['ref'] == variant.REF
    assert formated_variant['alt'] == variant.ALT[0]
    assert formated_variant['case_id'] == case_id
    assert formated_variant['homozygote'] == 0

def test_format_variant_no_gq(variant_no_gq, case_obj):
    ## GIVEN a variant without GQ
    variant = variant_no_gq
    case_id = case_obj['case_id']
    ## WHEN parsing the variant using a GQ treshold
    formated_variant = build_variant(
        variant=variant,
        case_obj=case_obj,
        case_id=case_id,
        gq_treshold=20
        )
    ## THEN assert that None is returned since requirements are not fulfilled
    assert formated_variant is None

def test_format_variant_chr_prefix(variant_chr, case_obj):
    ## GIVEN a variant with 'chr' prefix in chromosome name 
    variant = variant_chr
    assert variant.CHROM.startswith('chr')
    case_id = case_obj['case_id']
    ## WHEN parsing the variant using a GQ treshold
    formated_variant = build_variant(
        variant=variant,
        case_obj=case_obj,
        case_id=case_id,
        gq_treshold=20
        )
    ## THEN assert that the 'chr' part has been stripped away
    assert formated_variant['chrom'] == variant.CHROM[3:]


def test_format_variant_no_family_id(het_variant, case_obj):
    ## GIVEN a parsed variant
    variant = het_variant
    case_id = case_obj['case_id']
    ## WHEN parsing the variant telling that 'case_id' is None
    formated_variant = build_variant(
        variant=variant,
        case_obj=case_obj,
        case_id=None
        )
    ## THEN assert that case_id was not added
    assert formated_variant.get('case_id') == None
    assert formated_variant['homozygote'] == 0
    assert formated_variant['hemizygote'] == 0

def test_format_homozygote_variant(hom_variant, case_obj):
    ## GIVEN a parsed hom variant
    variant = hom_variant
    case_id = case_obj['case_id']

    ## WHEN parsing the variant
    formated_variant = build_variant(
        variant=variant,
        case_obj=case_obj,
        case_id=case_id
        )

    ## THEN assert that the variant has hom count
    assert formated_variant['homozygote'] == 1
    assert formated_variant['hemizygote'] == 0

def test_format_hemizygote_variant(hem_variant, case_obj):
    ## GIVEN a parsed hemizygous variant
    variant = hem_variant
    case_id = case_obj['case_id']

    ## WHEN parsing the variant
    formated_variant = build_variant(
        variant=variant,
        case_obj=case_obj,
        case_id=case_id
        )

    ## THEN assert that hemizygote count is 1
    assert formated_variant['homozygote'] == 0
    assert formated_variant['hemizygote'] == 1

def test_format_variant_no_call(variant_no_call, case_obj):
    ## GIVEN a parsed variant with no call in all individuals
    variant = variant_no_call
    case_id = case_obj['case_id']
    
    for call in variant.gt_types:
        assert GENOTYPE_MAP[call] in ['no_call', 'hom_ref']

    ## WHEN parsing the variant
    formated_variant = build_variant(
        variant=variant,
        case_obj=case_obj,
        case_id=case_id
        )

    ## THEN assert that the result is None
    assert formated_variant is None



import logging
from pprint import pprint as pp

from loqusdb.plugins import BaseCaseMixin
from loqusdb.exceptions import CaseError

LOG = logging.getLogger(__name__)

class CaseMixin(BaseCaseMixin):

    def case(self, case):
        """Get a case from the database

            Search the cases with the case id

            Args:
                case (dict): A case dictionary

            Returns:
                mongo_case (dict): A mongo case dictionary
        """
        LOG.debug("Getting case {0} from database".format(case.get('case_id')))
        case_id = case['case_id']
        return self.db.case.find_one({'case_id': case_id})

    def cases(self):
        """Get all cases from the database

            Returns:
                cases (Iterable(Case)): A iterable with mongo cases
        """
        LOG.debug("Collecting all cases from database")
        return self.db.case.find()

    def add_case(self, case, update=False):
        """Add a case to the case collection

        If the case exists and update is False raise error.

        Args:
            db (MongoClient): A connection to the mongodb
            case (dict): A case dictionary
            update(bool): If existing case should be updated

        Returns:
            mongo_case_id(ObjectId)

        """
        existing_case = self.case(case)
        if existing_case and not update:
            raise CaseError("Case {} already exists".format(case['case_id']))
        if existing_case:
            self.db.case.find_one_and_replace(
                {'case_id': case['case_id']},
                case,
            )
        else:
            self.db.case.insert_one(case)

        return case

    def delete_case(self, case):
        """Delete case from the database

            Delete a case from the database

            Args:
                case (dict): A case dictionary

        """
        mongo_case = self.case(case)

        if not mongo_case:
            raise CaseError("Tried to delete case {0} but could not find case".format(
                case.get('case_id')
            ))
        LOG.info("Removing case {0} from database".format(
            mongo_case.get('case_id')
        ))
        self.db.case.delete_one({'_id': mongo_case['_id']})

        return

    def case_count(self):
        """Returns the total number of cases in the database

        returns:
            nr_of_cases (int): Total number of cases in database
        """
        nr_of_cases = 0
        res = self.cases

        return res.count()



import logging
import click

from datetime import datetime

from loqusdb.utils.migrate import migrate_database
from . import base_command

LOG = logging.getLogger(__name__)


@base_command.command('migrate', short_help="Migrate an old loqusdb instance")
@click.pass_context
def migrate(ctx,):
    """Migrate an old loqusdb instance to 1.0
    """
    adapter = ctx.obj['adapter']
    
    start_time = datetime.now()
    
    nr_updated = migrate_database(adapter)
    
    LOG.info("All variants updated, time to complete migration: {}".format(
        datetime.now() - start_time))
    LOG.info("Nr variants that where updated: %s", nr_updated)


from loqusdb.utils.delete import delete
from loqusdb.utils.load import load_database

def test_delete_case(mongo_adapter, simple_case):
    ## GIVEN a mongoadapter with a inserted case
    db = mongo_adapter.db
    db.case.insert_one(simple_case)
    mongo_case = db.case.find_one()

    assert mongo_case
    ## WHEN deleting the case

    mongo_adapter.delete_case(simple_case)
    ## THEN assert that the case was deleted
    mongo_case = db.case.find_one()

    mongo_case = db.case.find_one()

    assert mongo_case == None



def test_delete_case_and_variants(vcf_path, ped_path, real_mongo_adapter, case_id, case_obj):
    mongo_adapter = real_mongo_adapter
    db = mongo_adapter.db

    load_database(
        adapter=mongo_adapter,
        variant_file=vcf_path,
        family_file=ped_path,
        family_type='ped',
    )

    mongo_case = db.case.find_one()

    assert mongo_case['case_id'] == case_id

    delete(
        adapter=mongo_adapter,
        case_obj=case_obj,
    )

    mongo_case = db.case.find_one()

    assert mongo_case == None

    mongo_variant = db.case.find_one()

    assert mongo_variant == None



import pytest
from pprint import pprint as pp
from loqusdb.exceptions import CaseError

from loqusdb.utils.load import (update_case,load_case)

def test_load_family(mongo_adapter, simple_case):
    ## GIVEN a adapter to an empty database
    db = mongo_adapter.db
    
    ## WHEN loading a family to the database
    mongo_adapter.add_case(simple_case)
    
    ## THEN assert that the case was loaded
    mongo_case = db.case.find_one()
    assert mongo_case['case_id'] == simple_case['case_id']


def test_load_same_family_twice(mongo_adapter, simple_case):
    ## GIVEN a adapter to an empty database
    db = mongo_adapter.db

    ## WHEN loading a family to the database twice
    mongo_adapter.add_case(simple_case)
    
    ## THEN assert that a CaseError is raised
    with pytest.raises(CaseError):
        mongo_adapter.add_case(simple_case)
        
def test_update_case(case_obj, sv_case_obj):
    ## GIVEN an existing case and a case with new info
    
    ## WHEN merging the two case objs
    updated_case = update_case(sv_case_obj, case_obj)
    
    ## THEN assert that the updated case includes information from both cases
    assert updated_case['sv_individuals'] == sv_case_obj['sv_individuals']
    assert updated_case['sv_individuals'] != case_obj['sv_individuals']

    assert updated_case['_sv_inds'] == sv_case_obj['_sv_inds']
    assert updated_case['_sv_inds'] != case_obj['_sv_inds']

    assert updated_case['nr_variants'] == case_obj['nr_variants']
    assert updated_case['nr_sv_variants'] == sv_case_obj['nr_sv_variants']

def test_update_case_same_vcf(case_obj, sv_case_obj):
    ## GIVEN an existing case and a case with new info
    sv_case_obj['vcf_path'] = case_obj['vcf_path']
    ## WHEN merging the two case objs
    ## THEN assert that a CaseError is raised since we are trying to modify existing VCF
    with pytest.raises(CaseError):
        updated_case = update_case(sv_case_obj, case_obj)

def test_load_complete_case(mongo_adapter, complete_case_obj):
    ## GIVEN a case that includes both svs and snvs
    db = mongo_adapter.db
    
    ## WHEN loading the case
    case_obj = load_case(mongo_adapter,complete_case_obj)
    ## THEN assert that all info is added
    loaded_case = db.case.find_one()
    
    assert len(loaded_case['individuals']) == 3
    assert len(loaded_case['sv_individuals']) == 3
    assert loaded_case['nr_variants'] > 0
    assert loaded_case['nr_sv_variants'] > 0
    


import pytest
import gzip

from pprint import pprint as pp

from loqusdb.utils.vcf import (get_vcf, get_file_handle, check_vcf)
from cyvcf2 import VCF

def test_get_file_handle(vcf_path):
    ## GIVEN the path to a vcf
    
    ## WHEN geting the file handle
    vcf = get_file_handle(vcf_path)
    
    ## THEN assert that a VCF object is returned
    assert type(vcf) is VCF

def test_get_zipped_file_handle(zipped_vcf_path):
    ## GIVEN the path to a zipped vcf

    ## WHEN geting the file handle
    vcf = get_file_handle(zipped_vcf_path)

    ## THEN assert that a VCF object is returned
    assert type(vcf) is VCF

def test_get_vcf_non_vcf(ped_path):
    ## GIVEN the path to a non vcf

    ## WHEN geting the file handle

    ## THEN assert that a IOError is raised
    with pytest.raises(IOError):
        vcf = get_file_handle(ped_path)

def test_get_vcf_non_existing():
    ## GIVEN the path to a non existing file

    ## WHEN geting the file handle

    ## THEN assert that a IOError is raised
    with pytest.raises(IOError):
        vcf = get_file_handle('hello')

def test_check_vcf(vcf_path):
    ## GIVEN the path to a vcf
    nr_variants = 0
    vcf = VCF(vcf_path)
    inds = vcf.samples
    for var in vcf:
        nr_variants += 1
    ## WHEN checking the vcf
    vcf_info = check_vcf(vcf_path)

    ## THEN assert that the number of variants is correct
    assert vcf_info['nr_variants'] == nr_variants

    ## THEN assert that the individuals are returned
    assert vcf_info['individuals'] == inds

    ## THEN assert that the variant type is correct
    assert vcf_info['variant_type'] == 'snv'
    



from loqusdb.plugins import MongoAdapter

def test_connect(mongo_client):
    db_name = 'test'
    adapter = MongoAdapter(mongo_client, db_name)
    
    assert adapter.db_name == db_name

# def test_connect_uri(mongo_client):
#     adapter = MongoAdapter()
#     uri = "{0}".format("mongodb://localhost")
#     adapter.connect(uri=uri)
#     assert True


import pytest

from loqusdb.utils.case import get_case
from loqusdb.exceptions import CaseError

def test_get_family(case_lines):
    family = get_case(case_lines)
    assert family.family_id == 'recessive_trio'

def test_get_multiple_families(two_cases):
    with pytest.raises(CaseError):
        family = get_case(two_cases)


import logging

from copy import deepcopy

from ped_parser import FamilyParser

from loqusdb.exceptions import CaseError

LOG = logging.getLogger(__name__)

def get_case(family_lines, family_type='ped', vcf_path=None):
    """Return ped_parser case from a family file
    
    Create a dictionary with case data. If no family file is given create from VCF
    
    Args:
        family_lines (iterator): The family lines
        family_type (str): The format of the family lines
        vcf_path(str): Path to VCF
    
    Returns:
        family (Family): A ped_parser family object
    """
    family = None
    LOG.info("Parsing family information")
    
    family_parser = FamilyParser(family_lines, family_type)
    
    families = list(family_parser.families.keys())
    
    LOG.info("Found families {0}".format(', '.join(families)))
    
    if len(families) > 1:
        raise CaseError("Only one family per load can be used")
    
    family = family_parser.families[families[0]]
    
    return family

def update_case(case_obj, existing_case):
    """Update an existing case
    
    This will add paths to VCF files, individuals etc
    
    Args:
        case_obj(models.Case)
        existing_case(models.Case)
    
    Returns:
        updated_case(models.Case): Updated existing case
    """
    variant_nrs = ['nr_variants', 'nr_sv_variants']
    individuals = [('individuals','_inds'), ('sv_individuals','_sv_inds')]
    
    updated_case = deepcopy(existing_case)
    
    for i,file_name in enumerate(['vcf_path','vcf_sv_path']):
        variant_type = 'snv'
        if file_name == 'vcf_sv_path':
            variant_type = 'sv'
        if case_obj.get(file_name):
            if updated_case.get(file_name):
                LOG.warning("VCF of type %s already exists in case", variant_type)
                raise CaseError("Can not replace VCF in existing case")
            else:
                updated_case[file_name] = case_obj[file_name]
                updated_case[variant_nrs[i]] = case_obj[variant_nrs[i]]
                updated_case[individuals[i][0]] = case_obj[individuals[i][0]]
                updated_case[individuals[i][1]] = case_obj[individuals[i][1]]

    return updated_case



# -*- coding: utf-8 -*-
from .dotdict import DotDict


class Variant(dict):
    """Represent a Variant.
    
    snv/indels and svs differ quite a lot, we are here representing them with the same object.
    snv/indels will miss some of the information
    """
    def __init__(self, chrom, pos, end, ref, alt, variant_id=None, end_chrom=None, sv_type=None,
                 sv_len=None, case_id=None, observations=0, homozygote=0, hemizygote=0, 
                 is_sv=False, id_column=None):
        super(Variant, self).__init__(
            _id=variant_id,
            variant_id=variant_id, 
            chrom=chrom, 
            pos=pos, 
            end=end, 
            ref=ref, 
            alt=alt,
            end_chrom=end_chrom, 
            sv_type = sv_type,
            sv_len = sv_len,
            case_id = case_id,
            observations = observations,
            homozygote = homozygote,
            hemizygote = hemizygote,
            is_sv=is_sv,
            id_column = id_column,
        )
    



from .cli import cli as base_command

from .wipe import wipe as wipe_command
from .load import load as load_command
from .update import update as update_command
from .delete import delete as delete_command
from .export import export as export_command
from .view import cases as cases_command
from .view import variants as variants_command
from .view import index as index_command
from .migrate import migrate as migration_command
from .identity import identity as identity_command



import logging
import click
import coloredlogs

from pprint import pprint as pp

from mongomock import MongoClient as MockClient

from pymongo import uri_parser

from mongo_adapter import get_client
from mongo_adapter.exceptions import Error as DB_Error

from loqusdb.log import LEVELS, init_log
from loqusdb import __version__
from loqusdb.plugins import MongoAdapter

LOG_LEVELS = ['DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL']

LOG = logging.getLogger(__name__)

@click.group()
@click.option('-db', '--database',
                default='loqusdb',
                show_default=True,
)
@click.option('-u', '--username',
                type=str
)
@click.option('-p', '--password',
                type=str
)
@click.option('-port', '--port',
                default=27017,
                show_default=True,
                help='Specify the port where to look for the mongo database.'
)
@click.option('-h', '--host',
                default='localhost',
                show_default=True,
                help='Specify the host where to look for the mongo database.'
)
@click.option('--uri',
                help='Specify a mongodb uri'
)
# @click.option('-b', '--backend',
#                 default='mongo',
#                 show_default=True,
#                 type=click.Choice(['mongo',]),
#                 help='Specify what backend to use.'
# )
@click.option('-t', '--test',
                is_flag=True,
                help='Used for testing.'
)
@click.option('-v', '--verbose', is_flag=True)
@click.version_option(__version__)
@click.pass_context
def cli(ctx, database, username, password, port, host, uri, verbose, test):
    """loqusdb: manage a local variant count database."""
    # configure root logger to print to STDERR
    loglevel = "INFO"
    if verbose:
        loglevel = "DEBUG"
    coloredlogs.install(level=loglevel)

    if test:
        uri = "mongomock://"
    try:
        client = get_client(
            host=host, 
            port=port, 
            username=username,
            password=password,
            uri=uri,
        )
    except DB_Error as err:
        LOG.warning(err)
        ctx.abort()

    adapter = MongoAdapter(client, db_name=database)

    ctx.obj = {}
    ctx.obj['db'] = database
    ctx.obj['user'] = username
    ctx.obj['password'] = password
    ctx.obj['port'] = port
    ctx.obj['host'] = host
    ctx.obj['adapter'] = adapter
    ctx.obj['version'] = __version__



import logging
import click
import tempfile

from datetime import datetime

from vcftoolbox import (HeaderParser, print_headers, print_variant)

from loqusdb import CHROMOSOME_ORDER
from loqusdb.utils.variant import format_variant

from . import base_command

LOG = logging.getLogger(__name__)

@base_command.command('export', short_help="Export variants to VCF format")
@click.option('-o', '--outfile',
    type=click.File('w'),
    help='Specify the path to a file where results should be stored.'
)
@click.option('-t','--variant-type',
    type=click.Choice(['sv','snv']),
    default='snv',
    show_default=True,
    help="If svs or snvs should be exported"
)
@click.pass_context
def export(ctx, outfile, variant_type):
    """Export the variants of a loqus db
        
        The variants are exported to a vcf file
    """
    adapter = ctx.obj['adapter']
    version = ctx.obj['version']
    
    LOG.info("Export the variants from {0}".format(adapter))
    nr_cases = 0

    is_sv = variant_type == 'sv'
    existing_chromosomes = set(adapter.get_chromosomes(sv=is_sv))
    
    ordered_chromosomes = []
    for chrom in CHROMOSOME_ORDER:
        if chrom in existing_chromosomes:
            ordered_chromosomes.append(chrom)
            existing_chromosomes.remove(chrom)
    for chrom in existing_chromosomes:
        ordered_chromosomes.append(chrom)
    
    nr_cases = adapter.cases().count()
    LOG.info("Found {0} cases in database".format(nr_cases))

    head = HeaderParser()
    head.add_fileformat("VCFv4.3")
    head.add_meta_line("NrCases", nr_cases)
    head.add_info("Obs", '1', 'Integer', "The number of observations for the variant")
    head.add_info("Hom", '1', 'Integer', "The number of observed homozygotes")
    head.add_info("Hem", '1', 'Integer', "The number of observed hemizygotes")
    head.add_version_tracking("loqusdb", version, datetime.now().strftime("%Y-%m-%d %H:%M"))
    
    if variant_type == 'sv':
        head.add_info("END", '1', 'Integer', "End position of the variant")
        head.add_info("SVTYPE", '1', 'String', "Type of structural variant")
        head.add_info("SVLEN", '1', 'Integer', "Length of structural variant")
        
        
    for chrom in ordered_chromosomes:
        length = adapter.get_max_position(chrom)
        head.add_contig(contig_id=chrom, length=str(length))

    print_headers(head, outfile=outfile)
    
    for chrom in ordered_chromosomes:
        if variant_type == 'snv':
            LOG.info("Collecting all SNV variants")
            variants = adapter.get_variants(chromosome=chrom)
        else:
            LOG.info("Collecting all SV variants")
            variants = adapter.get_sv_variants(chromosome=chrom)
        LOG.info("{} variants found".format(variants.count()))
        for variant in variants:
            variant_line = format_variant(variant, variant_type=variant_type)
            # chrom = variant['chrom']
            # pos = variant['start']
            # ref = variant['ref']
            # alt = variant['alt']
            # observations = variant['observations']
            # homozygotes = variant['homozygote']
            # hemizygotes = variant['hemizygote']
            # info = "Obs={0}".format(observations)
            # if homozygotes:
            #     info += ";Hom={0}".format(homozygotes)
            # if hemizygotes:
            #     info += ";Hem={0}".format(hemizygotes)
            # variant_line = "{0}\t{1}\t.\t{2}\t{3}\t.\t.\t{4}\n".format(
            #     chrom, pos, ref, alt, info)
            print_variant(variant_line=variant_line, outfile=outfile)



import logging

from pprint import pprint as pp
from collections import namedtuple

from loqusdb.models import Variant 
from loqusdb.exceptions import CaseError
from loqusdb.constants import (PAR,GENOTYPE_MAP,CHROM_TO_INT)

LOG = logging.getLogger(__name__)

Position = namedtuple('Position', 'chrom pos')

# These are coordinate for the pseudo autosomal regions in GRCh37

def check_par(chrom, pos):
    """Check if a coordinate is in the PAR region
    
        Args:
            chrom(str)
            pos(int)
    
        Returns:
            par(bool)
    """
    par = False
    
    for interval in PAR.get(chrom,[]):
        if (pos >= interval[0] and pos <= interval[1]):
            par = True
    
    return par
    

def get_variant_id(variant):
    """Get a variant id on the format chrom_pos_ref_alt"""
    variant_id = '_'.join([
            str(variant.CHROM),
            str(variant.POS),
            str(variant.REF),
            str(variant.ALT[0])
        ]
    )
    return variant_id

def is_greater(a,b):
    """Check if position a is greater than position b
    This will look at chromosome and position.
    
    For example a position where chrom = 2 and pos = 300 is greater than a position where
    chrom = 1 and pos = 1000
    
    If any of the chromosomes is outside [1-22,X,Y,MT] we can not say which is biggest.
    
    Args:
        a,b(Position)
    
    Returns:
        bool: True if a is greater than b
    """
    
    a_chrom = CHROM_TO_INT.get(a.chrom,0)
    b_chrom = CHROM_TO_INT.get(b.chrom,0)
    
    if (a_chrom == 0 or b_chrom == 0):
        return False
    
    if a_chrom > b_chrom:
        return True
    
    if a_chrom == b_chrom:
        if a.pos > b.pos:
            return True
    
    return False


def build_variant(variant, case_obj, case_id=None, gq_treshold=None):
    """Return a Variant object

    Take a cyvcf2 formated variant line and return a models.Variant.

    If criterias are not fullfilled, eg. variant have no gt call or quality
    is below gq treshold then return None.

    Args:
        variant(cyvcf2.Variant)
        case_obj(Case): We need the case object to check individuals sex
        case_id(str): The case id
        gq_treshold(int): Genotype Quality treshold
    
    Return:
        formated_variant(models.Variant): A variant dictionary
    """
    variant_obj = None

    sv = False
    # Let cyvcf2 tell if it is a Structural Variant or not
    if variant.var_type == 'sv':
        sv = True

    chrom = variant.CHROM
    if chrom.startswith(('chr', 'CHR', 'Chr')):
        chrom = chrom[3:]
    
    pos = int(variant.POS)
    
    # Get the end position
    # This will be None for non-svs
    end_pos = variant.INFO.get('END')
    if end_pos:
        end = int(end_pos)
    else:
        end = int(variant.end)

    # chrom_pos_ref_alt
    variant_id = get_variant_id(variant)

    ref = variant.REF
    # ALT is an array in cyvcf2
    alt = variant.ALT[0]

    # Default end chrom is chrom
    end_chrom = chrom

    sv_type = variant.INFO.get('SVTYPE')
    length = variant.INFO.get('SVLEN')
    if length:
        sv_len = abs(length)
    else:
        sv_len = end - pos

    # Translocations will sometimes have a end chrom that differs from chrom
    if sv_type == 'BND':
        other_coordinates = alt.strip('ACGTN[]').split(':')
        end_chrom = other_coordinates[0]
        if end_chrom.startswith(('chr', 'CHR', 'Chr')):
            end_chrom = end_chrom[3:]

        end = int(other_coordinates[1])

        #Set 'infinity' to length if translocation
        sv_len = float('inf')
        sv_type = 'BND'

    # Insertions often have length 0 in VCF
    if (sv_len == 0 and alt != '<INS>'):
        sv_len = len(alt)

    if (pos == end) and (sv_len > 0):
        end = pos + sv_len

    position = Position(chrom, pos)
    end_position = Position(end_chrom, end)
    
    # If 'start' is greater than 'end', switch positions
    if is_greater(position, end_position):
        end_chrom = position.chrom
        end = position.pos
        
        chrom = end_position.chrom
        pos = end_position.pos
        

    # These are integers that will be used when uploading
    found_homozygote = 0
    found_hemizygote = 0

    # Only look at genotypes for the present individuals
    if sv:
        found_variant = True
    else:
        found_variant = False
        for ind_obj in case_obj['individuals']:
            ind_id = ind_obj['ind_id']
            # Get the index position for the individual in the VCF
            ind_pos = ind_obj['ind_index']
            gq = int(variant.gt_quals[ind_pos])
            if (gq_treshold and gq < gq_treshold):
                continue

            genotype = GENOTYPE_MAP[variant.gt_types[ind_pos]]

            if genotype in ['het', 'hom_alt']:
                LOG.debug("Found variant")
                found_variant = True

                # If variant in X or Y and individual is male,
                # we need to check hemizygosity
                if chrom in ['X','Y'] and ind_obj['sex'] == 1:
                    if not check_par(chrom, pos):
                        LOG.debug("Found hemizygous variant")
                        found_hemizygote = 1

                if genotype == 'hom_alt':
                    LOG.debug("Found homozygote alternative variant")
                    found_homozygote = 1

    if found_variant:
        variant_obj = Variant(
            variant_id=variant_id,
            chrom=chrom,
            pos=pos,
            end=end,
            ref=ref,
            alt=alt,
            end_chrom=end_chrom,
            sv_type = sv_type,
            sv_len = sv_len,
            case_id = case_id,
            homozygote = found_homozygote,
            hemizygote = found_hemizygote,
            is_sv = sv,
            id_column = variant.ID,
        )

    return variant_obj


# -*- coding: utf-8 -*-

class Case(dict):
    """Represent a Case."""
    def __init__(self, case_id, vcf_path=None, vcfsv_path=None, nr_variants=None, nr_sv_variants=None):
        super(Case, self).__init__(
            case_id=case_id,
            vcf_path=vcf_path,
            vcf_sv_path=vcfsv_path,
            nr_variants=nr_variants,
            nr_sv_variants=nr_sv_variants,
        )
        self['individuals'] = []
        self['sv_individuals'] = []
        self['_inds'] = {}
        self['_sv_inds'] = {}


class Individual(dict):
    """Individual representation."""
    def __init__(self, ind_id, case_id=None, mother=None,
                 father=None, sex=None, phenotype=None, ind_index=None):
        """Construct a individual object

            Args:
                ind_id (str): The individual id
                case_id (str): What case it belongs to
                mother (str): The mother id
                father (str): The father id
                sex (str): Sex in ped format
                phenotype (str): Phenotype in ped format
                ind_index (int): Column in the vcf.
        """
        super(Individual, self).__init__(
            ind_id=ind_id,
            name=ind_id,
            case_id=case_id,
            ind_index=ind_index,
            sex=sex,
        )



import logging
from pkg_resources import get_distribution

from pymongo import (IndexModel, ASCENDING, DESCENDING)

logger = logging.getLogger(__name__)

__version__ = get_distribution("loqusdb").version

INDEXES = {
    'variant' : [
        IndexModel([
        ('chrom', ASCENDING),
        ('pos', ASCENDING),
        ('end', ASCENDING),
        ], name="coordinates"),
        
        IndexModel([
        ('chrom', ASCENDING),
        ('end', ASCENDING),
        ], name="end"),
        
    ],
    'structural_variant' : [
        IndexModel([
        ('chrom', ASCENDING),
        ('end_chrom', ASCENDING),
        ('sv_type', ASCENDING),
        ('pos_left', ASCENDING),
        ], name="coordinates"),
        
        IndexModel([
        ('chrom', ASCENDING),
        ('pos_left', ASCENDING),
        ('end_right', ASCENDING),
        ], name="short_coordinates"),
    ],
    'identity' : [
        IndexModel([
        ('cluster_id', ASCENDING),
        ], name="cluster"),
        
        IndexModel([
        ('variant_id', ASCENDING),
        ], name="variant"),
    ],
    
}

CHROMOSOME_ORDER = ('1', '2', '3', '4', '5', '6', '7', '8', '9', '10', '11', '12', 
               '13', '14', '15', '16', '17', '18', '19', '20', '21', '22', 
               '23', 'X', 'Y', 'MT')


import os
import logging

from cyvcf2 import VCF

from loqusdb.exceptions import VcfError
from loqusdb.build_models.variant import get_variant_id

LOG = logging.getLogger(__name__)
VALID_ENDINGS = ['.vcf', '.gz']

def get_file_handle(file_path):
    """Return cyvcf2 VCF object
    
    Args:
        file_path(str)
    
    Returns:
        vcf_obj(cyvcf2.VCF)
    """
    LOG.debug("Check if file end is correct")
    
    if not os.path.exists(file_path):
        raise IOError("No such file:{0}".format(file_path))
    
    if not os.path.splitext(file_path)[-1] in VALID_ENDINGS:
        raise IOError("Not a valid vcf file name: {}".format(file_path))
    
    vcf_obj = VCF(file_path)
    
    return vcf_obj

def get_vcf(file_path):
    """Yield variants from a vcf file
    
    Args:
        file_path(str)
    
    Yields:
        vcf_obj(cyvcf2.VCF): An iterable with cyvcf2.Variant

    """

    vcf_obj = get_file_handle(file_path)
    
    return vcf_obj

def check_sorting(previous_chrom, previous_pos, current_chrom, current_pos):
    """docstring for check_sorting"""
    pass

def check_vcf(vcf_path, expected_type='snv'):
    """Check if there are any problems with the vcf file

    Args:
        vcf_path(str)
        expected_type(str): 'sv' or 'snv'

    Returns:
        vcf_info(dict): dict like 
        {
            'nr_variants':<INT>, 
            'variant_type': <STR> in ['snv', 'sv'],
            'individuals': <LIST> individual positions in file
        }
    """
    LOG.info("Check if vcf is on correct format...")

    vcf = VCF(vcf_path)
    individuals = vcf.samples
    variant_type = None

    previous_pos = None
    previous_chrom = None

    posititon_variants = set()

    nr_variants = 0
    for nr_variants,variant in enumerate(vcf,1):

        # Check the type of variant
        current_type = 'sv' if variant.var_type == 'sv' else 'snv'
        if not variant_type:
            variant_type = current_type

        # Vcf can not include both snvs and svs
        if variant_type != current_type:
            raise VcfError("Vcf includes a mix of snvs and svs")

        current_chrom = variant.CHROM
        current_pos = variant.POS
        
        # We start with a simple id that can be used by SV:s
        variant_id = "{0}_{1}".format(current_chrom, current_pos)
        # For SNVs we can create a proper variant id with chrom_pos_ref_alt
        if variant_type == 'snv':
            variant_id = get_variant_id(variant)

        # Initiate variables
        if not previous_chrom:
            previous_chrom = current_chrom
            previous_pos = current_pos
            posititon_variants = set([variant_id])
            continue

        # Update variables if new chromosome
        if current_chrom != previous_chrom:
            previous_chrom = current_chrom
            previous_pos = current_pos
            posititon_variants = set([variant_id])
            continue
        
        if variant_type == 'snv':
            # Check if variant is unique
            if current_pos == previous_pos:
                if variant_id in posititon_variants:
                    raise VcfError("Variant {0} occurs several times"\
                                   " in vcf".format(variant_id))
                else:
                    posititon_variants.add(variant_id)
            # Check if vcf is sorted
            else:
                if not current_pos >= previous_pos:
                    raise VcfError("Vcf if not sorted in a correct way")
                previous_pos = current_pos
                # Reset posititon_variants since we are on a new position
                posititon_variants = set([variant_id])

    if variant_type != expected_type:
        raise VcfError("VCF file does not only include {0}s, please check vcf {1}".format(
                        expected_type.upper(), vcf_path))

    LOG.info("Vcf file %s looks fine", vcf_path)
    LOG.info("Nr of variants in vcf: {0}".format(nr_variants))
    LOG.info("Type of variants in vcf: {0}".format(variant_type))
    
    vcf_info = {
        'nr_variants': nr_variants,
        'variant_type': variant_type,
        'individuals': individuals,
    }

    return vcf_info



# -*- coding: utf-8 -*-
import logging
import click
from datetime import datetime

from loqusdb.exceptions import CaseError
from loqusdb.utils.delete import delete as delete_command
from . import base_command

LOG = logging.getLogger(__name__)

@base_command.command('delete', short_help="Delete the variants of a family")
@click.argument('variant-file',
                    type=click.Path(exists=True),
                    metavar='<vcf_file>'
)
@click.option('-f', '--family-file',
                    type=click.Path(exists=True),
                    metavar='<ped_file>'
)
@click.option('-t' ,'--family-type', 
                type=click.Choice(['ped', 'alt', 'cmms', 'mip']), 
                default='ped',
                help='If the analysis use one of the known setups, please specify which one.'
)
@click.option('-c' ,'--case-id', 
                type=str, 
                help='If a different case id than the one in ped file should be used'
)
@click.pass_context
def delete(ctx, variant_file, family_file, family_type, case_id):
    """Delete the variants of a case."""
    if not (family_file or case_id):
        LOG.error("Please provide a family file")
        ctx.abort()
    
    adapter = ctx.obj['adapter']

    start_deleting = datetime.now()
    try:
        delete_command(
            adapter=adapter, 
            variant_file=variant_file, 
            family_file=family_file, 
            family_type=family_type,
            case_id=case_id
        )
    except (CaseError, IOError) as error:
        LOG.warning(error)
        ctx.abort()



from loqusdb.utils.delete import delete_variants
from loqusdb.utils.load import load_variants
from loqusdb.build_models.variant import get_variant_id


def test_delete_variants(mongo_adapter, het_variant, case_obj):
    ## GIVEN a database with one variant
    db = mongo_adapter.db
    case_id = case_obj['case_id']
    
    db.variant.insert_one({
        '_id': get_variant_id(het_variant),
        'families': [case_id],
        'observations': 1,
    })
    
    mongo_variant = db.variant.find_one()
    assert mongo_variant['families'] == [case_id]

    ## WHEN deleting the variant
    delete_variants(
        adapter=mongo_adapter,
        vcf_obj=[het_variant],
        case_obj=case_obj
    )

    mongo_variant = db.variant.find_one()

    ## THEN assert that the variant was not found
    assert mongo_variant == None

def test_delete_variant(mongo_adapter, het_variant, case_obj):
    ## GIVEN a database with one variant that is observed twice
    db = mongo_adapter.db
    case_id = case_obj['case_id']
    
    db.variant.insert_one({
        '_id': get_variant_id(het_variant),
        'families': [case_id, '2'],
        'observations': 2,
    })

    mongo_variant = db.variant.find_one()
    assert mongo_variant['observations'] == 2

    ## WHEN deleting the variant for one case
    delete_variants(
        adapter=mongo_adapter,
        vcf_obj=[het_variant],
        case_obj=case_obj,
        case_id='2',
    )

    mongo_variant = db.variant.find_one()

    ## THEN assert that one case has been removed from 'families'
    assert mongo_variant['families'] == [case_id]
    ## THEN assert that the observation count is decreased
    assert mongo_variant['observations'] == 1

def test_delete_non_existing_variant(mongo_adapter, het_variant, case_obj):
    """docstring for test_load_variants"""
    ## GIVEN a mongo adapter to an empty database
    db = mongo_adapter.db
    case_id = case_obj['case_id']

    ## WHEN deleting the variants
    delete_variants(
        adapter=mongo_adapter,
        vcf_obj=[het_variant],
        case_obj=case_obj
    )

    # THEN assert nothing happens
    mongo_variant = db.variant.find_one()

    assert mongo_variant == None



from loqusdb.plugins import MongoAdapter
from mongomock import MongoClient

class MockFlaskApp(object):
    """Mock a Flask App"""
    def __init__(self, host='localhost', port=27017, db_name='test'):
        client = MongoClient()
        
        self.config = {}
        self.extensions = {'pymongo': {}}
        

        self.config['MONGO_DBNAME'] = db_name
        self.config['MONGO_PORT'] = port
        self.config['MONGO_HOST'] = host
        
        # This is how flaskpymongo sets it up:
        self.extensions['pymongo']['MONGO'] = [client, client[db_name]]

def test_init_app(mongo_client):
    app = MockFlaskApp()
    adapter = MongoAdapter()
    adapter.init_app(app)
    assert adapter.db_name == app.config['MONGO_DBNAME']


from loqusdb.constants import CHROM_TO_INT
from loqusdb.build_models.variant import (Position,is_greater)

def test_is_greater_different_chrom():
    ## GIVEN two positions where a is smaller than b. Different chroms
    a = Position('1',100)
    b = Position('3',100)
    
    ## WHEN testing if a is greater than b
    res = is_greater(a,b)
    ## THEN assert a was not greater than b
    assert res is False
    
    ## WHEN testing if b is greater than a
    res = is_greater(b,a)
    ## THEN assert b was greater than a
    assert res is True

def test_is_greater_same_chrom():
    ## GIVEN two positions where a is smaller than b. Same chroms
    a = Position('1',100)
    b = Position('1',300)
    
    ## WHEN testing if a is greater than b
    res = is_greater(a,b)
    ## THEN assert a was not greater than b
    assert res is False
    
    ## WHEN testing if b is greater than a
    res = is_greater(b,a)
    ## THEN assert b was greater than a
    assert res is True

def test_is_greater_wierd_chrom():
    ## GIVEN two positions where one chrom is unknown b
    a = Position('1',100)
    b = Position('GRLrs2',300)
    
    ## WHEN testing if a is greater than b
    res = is_greater(a,b)
    ## THEN assert a was not greater than b
    assert res is False
    
    ## WHEN testing if b is greater than a
    res = is_greater(b,a)
    ## THEN assert a was not greater than b
    assert res is False


from pprint import pprint as pp
from loqusdb.build_models.variant import build_variant

def test_load_insertion(small_insert_variant, mongo_adapter, case_obj):
    adapter = mongo_adapter
    ## GIVEN a mongo adapter with a case
    variant = small_insert_variant
    case_id = case_obj['case_id']
    adapter.add_case(case_obj)
    
    ## WHEN loading a small insertion
    formated_variant = build_variant(
        variant=variant,
        case_obj=case_obj,
        case_id=case_id
    )
    adapter.add_structural_variant(formated_variant)
    
    ## THEN assert the object returned is correct
    variant_cluster = adapter.db.structural_variant.find_one()
    
    assert variant_cluster['families'] == [case_id]

def test_load_same_insertion_twice(small_insert_variant, mongo_adapter, case_obj):
    adapter = mongo_adapter
    ## GIVEN a mongo adapter with a case
    variant = small_insert_variant
    case_id = case_obj['case_id']
    adapter.add_case(case_obj)
    
    ## WHEN loading a small insertion
    formated_variant = build_variant(
        variant=variant,
        case_obj=case_obj,
        case_id=case_id
    )
    adapter.add_structural_variant(formated_variant)
    formated_variant['case_id'] = '2'
    adapter.add_structural_variant(formated_variant)
    
    ## THEN assert the object returned is correct
    variant_cluster = adapter.db.structural_variant.find_one()
    
    assert set(variant_cluster['families']) == set([case_id, '2'])

def test_load_translocation(translocation_variant, case_obj, mongo_adapter):
    adapter = mongo_adapter
    ## GIVEN a mongo adapter with a case
    variant = translocation_variant
    case_id = case_obj['case_id']
    adapter.add_case(case_obj)
    
    ## WHEN loading a small insertion
    formated_variant = build_variant(
        variant=variant,
        case_obj=case_obj,
        case_id=case_id
    )
    adapter.add_structural_variant(formated_variant)
    
    ## THEN assert the object returned is correct
    variant_cluster = adapter.db.structural_variant.find_one()

    assert variant_cluster['families'] == [case_id]

def test_load_same_translocation_twice(translocation_variant, case_obj, mongo_adapter):
    adapter = mongo_adapter
    ## GIVEN a mongo adapter with a case
    variant = translocation_variant
    case_id = case_obj['case_id']
    adapter.add_case(case_obj)
    
    ## WHEN loading a small insertion
    formated_variant = build_variant(
        variant=variant,
        case_obj=case_obj,
        case_id=case_id
    )
    adapter.add_structural_variant(formated_variant)
    
    formated_variant['case_id'] = '2'
    adapter.add_structural_variant(formated_variant)
    
    ## THEN assert the object returned is correct
    variant_cluster = adapter.db.structural_variant.find_one()
    
    assert set(variant_cluster['families']) == set([case_id, '2'])



import logging
from click import progressbar

LOG = logging.getLogger(__name__)


def migrate_database(adapter):
    """Migrate an old loqusdb instance to 1.0
    
    Args:
        adapter
    
    Returns:
        nr_updated(int): Number of variants that where updated
    """
    
    all_variants = adapter.get_variants()
    nr_variants = all_variants.count()
    nr_updated = 0
    with progressbar(all_variants, label="Updating variants", length=nr_variants) as bar:
        for variant in bar:
            # Do not update if the variants have the correct format
            if 'chrom' in variant:
                continue
            nr_updated += 1
            splitted_id = variant['_id'].split('_')
            
            chrom = splitted_id[0]
            start = int(splitted_id[1])
            ref = splitted_id[2]
            alt = splitted_id[3]
            
            # Calculate end
            end = start + (max(len(ref), len(alt)) - 1)
            
            adapter.db.variant.find_one_and_update(
                {'_id': variant['_id']},
                {
                    '$set': {
                        'chrom': chrom,
                        'start': start,
                        'end': end
                    }
                }
            )
    
    return nr_updated



import logging

from pprint import pprint as pp

from loqusdb.plugins import BaseVariantMixin
from loqusdb.models import Identity

from pymongo import (ASCENDING, DESCENDING, UpdateOne)

LOG = logging.getLogger(__name__)

class SVMixin():

    def add_structural_variant(self, variant, max_window = 3000):
        """Add a variant to the structural variants collection
        
        The process of adding an SV variant differs quite a bit from the 
        more straight forward case of SNV/INDEL.
        
        Variants are represented in the database by clusters that are two intervals,
        one interval for start(pos) and one for end. The size of the intervals changes
        according to the size of the variants. The maximum window size is a parameter.
        
        Here we need to search if the variant matches any of the existing
        clusters. Then we need to choose the closest cluster and update
        the boundaries for that cluster.
        
        Args:
            variant (dict): A variant dictionary
            max_window(int): Specify the maximum window size for large svs
        
        
                                        -
                                      -   -
                                    -       -
                                  -           -
                                -               -
                              -                   -
                           |-----|              |-----|
                          /\                         /\
                          |                          |
                pos - interval_size             end + interval_size
        """
        # This will return the cluster most similar to variant or None
        cluster = self.get_structural_variant(variant)
        # If there was no matcing cluster we need to create a new cluster
        if cluster is None:
            # The cluster will be populated with information later.
            cluster = {
                'chrom': variant['chrom'],
                'end_chrom': variant['end_chrom'],
                'sv_type': variant['sv_type'],
                'pos_sum': 0,
                'end_sum': 0,
                'observations': 0,
                'length': 0,
                'sv_type': variant['sv_type'],
                'families': [],
                
            }
            # Insert variant to get a _id
            _id = self.db.structural_variant.insert_one(cluster).inserted_id
            
            cluster['_id'] = _id
        
        case_id = variant.get('case_id')
        if case_id:
            # If the variant is already added for this case we continue
            # One case will only give duplicated information
            if case_id in cluster['families']:
                return
            else:
                # Insert the new case in the beginning of array
                cluster['families'].insert(0,case_id)
                # Make sure array does not grow out of bounds
                cluster['families'] = cluster['families'][:50]

        # Update number of times we have seen the event
        nr_events = cluster['observations'] + 1
        
        #                             -
        #                           -   -
        #                         -       -
        #                       -           -
        #                     -               -
        #                   -                   -
        #                |--.--|              |--.--|
    
        # This indicates the center for each of the end points for the event AFTER the new variant
        # is added to the cluster
        # i.e. the dots in the picture above
        pos_mean = int((cluster['pos_sum'] + variant['pos']) // (nr_events))
        end_mean = int((cluster['end_sum'] + variant['end']) // (nr_events))
        
        # We need to calculate the new cluster length
        # Handle translocation as a special case
        if cluster['sv_type'] != 'BND':
            cluster_len = end_mean - pos_mean
            # We need to adapt the interval size depending on the size of the cluster
            divider = 10
            if cluster_len < 1000:
                # We allow intervals for smaller variants to be relatively larger
                divider = 2
            elif cluster_len < 1000:
                # We allow intervals for smaller variants to be relatively larger
                divider = 5
            interval_size = int(min(round(cluster_len/divider, -2), max_window))
        else:
            # We need to treat translocations as a special case.
            # Set length to a huge number that mongodb can handle, float('inf') would not work.
            cluster_len = 10e10
            # This number seems large, if compared with SV size it is fairly small.
            interval_size = max_window * 2
        
        # If the length of SV is shorter than 500 the variant 
        # is considered precise
        # Otherwise the interval size is closest whole 100 number
        res = self.db.structural_variant.find_one_and_update(
            {'_id': cluster['_id']},
            {
                '$inc': {
                    'observations': 1,
                    'pos_sum': variant['pos'],
                    'end_sum': variant['end'],
                },
            
                '$set': {
                    'pos_left': max(pos_mean - interval_size, 0),
                    'pos_right': pos_mean + interval_size,
                    'end_left': max(end_mean - interval_size, 0),
                    'end_right': end_mean + interval_size,
                    'families': cluster['families'],
                    'length': cluster_len,
                }
            }
        )
        
        # Insert an identity object to link cases to variants and clusters
        identity_obj = Identity(cluster_id=cluster['_id'], variant_id=variant['id_column'], 
                                case_id=case_id)
        self.db.identity.insert_one(identity_obj)
        
        return

    def get_structural_variant(self, variant):
        """Check if there are any overlapping sv clusters

       Search the sv variants with chrom start end_chrom end and sv_type

       Args:
           variant (dict): A variant dictionary

       Returns:
           variant (dict): A variant dictionary
        """
        # Create a query for the database
        # This will include more variants than we want
        # The rest of the calculations will be done in python
        query = {
                'chrom': variant['chrom'],
                'end_chrom': variant['end_chrom'],
                'sv_type': variant['sv_type'],
                '$and': [
                    {'pos_left': {'$lte': variant['pos']}},
                    {'pos_right': {'$gte': variant['pos']}},
                ]
            }

        res = self.db.structural_variant.find(query).sort('pos_left',1)
        match = None
        distance = None
        closest_hit = None
        # First we check that the coordinates are correct
        # Then we count the distance to mean on both ends to see which variant is closest
        for hit in res:
            # We know from the query that the variants position is larger than the left most part of
            # the cluster. 
            # If the right most part of the cluster is smaller than the variant position they do
            # not overlap
            if hit['end_left'] > variant['end']:
                continue
            if hit['end_right'] < variant['end']:
                continue

            # We need to calculate the distance to see what cluster that was closest to the variant
            distance = (abs(variant['pos'] - (hit['pos_left'] + hit['pos_right'])/2) + 
                        abs(variant['end'] - (hit['end_left'] + hit['end_right'])/2))

            # If we have no cluster yet we set the curent to be the hit
            if closest_hit is None:
                match = hit
                closest_hit = distance
                continue

            # If the distance is closer than previous we choose current cluster
            if distance < closest_hit:
                # Set match to the current closest hit
                match = hit
                # Update the closest distance
                closest_hit = distance

        return match

    def get_sv_variants(self, chromosome=None, end_chromosome=None, sv_type=None, 
                        pos=None, end=None):
        """Return all structural variants in the database

        Args:
            chromosome (str)
            end_chromosome (str)
            sv_type (str)
            pos (int): Left position of SV
            end (int): Right position of SV

        Returns:
            variants (Iterable(Variant))
        """
        query = {}
        
        if chromosome:
            query['chrom'] = chromosome
        if end_chromosome:
            query['end_chrom'] = end_chromosome
        if sv_type:
            query['sv_type'] = sv_type
        if pos:
            if not '$and' in query:
                query['$and'] = []
            query['$and'].append({'pos_left': {'$lte': pos}})
            query['$and'].append({'pos_right': {'$gte': pos}})
            
        if end:
            if not '$and' in query:
                query['$and'] = []
            query['$and'].append({'end_left': {'$lte': end}})
            query['$and'].append({'end_right': {'$gte': end}})
        
        LOG.info("Find all sv variants {}".format(query))
        
        return self.db.structural_variant.find(query).sort([('chrom', ASCENDING), ('pos_left', ASCENDING)])

    def get_clusters(self, variant_id):
        """Search what clusters a variant belongs to
        
        Args:
            variant_id(str): From ID column in vcf
        
        Returns:
            clusters()
        """
        query = {'variant_id':variant_id}
        identities = self.db.identity.find(query)
        return identities



from pprint import pprint as pp

from loqusdb.commands import base_command

from click.testing import CliRunner

def test_base_command():
    runner = CliRunner()
    result = runner.invoke(base_command, [])

    assert result.exit_code == 0


def test_load_command(vcf_path, ped_path, real_mongo_adapter):
    ## GIVEN a vcf_path a ped_path
    runner = CliRunner()
    ## WHEN inserting a case via the CLI
    command = ['--database', 'test', 'load', '--variant-file', vcf_path, '-f', ped_path]
    result = runner.invoke(base_command, command)
    ## THEN assert that the cli exits without problems
    assert result.exit_code == 0

def test_load_command_no_ped(vcf_path):
    runner = CliRunner()
    command = ['--database', 'test', 'load', '--variant-file', vcf_path]
    result = runner.invoke(base_command, command)

    assert result.exit_code == 1

# def test_delete_command(mongo_client, vcf_path, ped_path):
#     runner = CliRunner()
#
#     load_command = ['--test', '--database', 'test', 'load', vcf_path, '-f', ped_path]
#     result = runner.invoke(base_command, load_command)
#
#     delete_command = ['--test', '--database', 'test', 'delete', vcf_path, '-f', ped_path]
#     result = runner.invoke(base_command, delete_command)
#
#     assert result.exit_code == 0


# def test_wipe_command(mongo_client, vcf_path, ped_path):
#     runner = CliRunner()
#
#     load_command = ['-c', 'mongomock://', '--database', 'test', 'load', vcf_path, '-f', ped_path]
#     runner.invoke(base_command, load_command)
#
#     wipe_command = ['-c', 'mongomock://', '--database', 'test', 'wipe', '--yes']
#     result = runner.invoke(base_command, wipe_command)
#
#     assert result.exit_code == 0




from loqusdb.utils.migrate import migrate_database

def test_migrate_one_variant(mongo_adapter):
    ##GIVEN a adapter to an empty old database
    adapter = mongo_adapter
    assert adapter.get_variants().count() == 0
    
    ##WHEN inserting a old style variant and updating the database
    chrom = '1'
    pos = 880086
    ref = 'T'
    alt = 'C'
    hom = hem = 0
    obs = 3
    families = ['recessive_trio', '1', '2']
    variant = {
        '_id': '_'.join([chrom, str(pos), ref, alt]),
        'homozygote': hom,
        'hemizygote': hem,
        'observations': obs,
        'families': families,
    }
    adapter.db.variant.insert_one(variant)
    ##THEN assert that the variant was updated correct
    
    nr_updated = migrate_database(adapter)
    assert nr_updated == 1
    
    migrated_variant = adapter.db.variant.find_one()
    
    assert migrated_variant['chrom'] == chrom
    assert migrated_variant['start'] == pos
    assert migrated_variant['end'] == pos
    assert migrated_variant['homozygote'] == hom
    assert migrated_variant['hemizygote'] == hem
    assert migrated_variant['observations'] == obs
    assert migrated_variant['families'] == families


from pprint import pprint as pp
from loqusdb.build_models import build_variant

def test_format_indel(del_variant, case_obj):
    ## GIVEN a SV deletion
    variant = del_variant
    case_id = case_obj['case_id']
    ## WHEN parsing the variant
    formated_variant = build_variant(
        variant=variant,
        case_obj=case_obj,
        case_id=case_id
        )
    expected_id = '_'.join([
        variant.CHROM,
        str(variant.POS), 
        variant.REF, 
        variant.ALT[0]
    ])
    
    ## THEN assert the sv is parsed correct
    assert formated_variant
    assert formated_variant['variant_id'] == expected_id
    assert formated_variant['chrom'] == variant.CHROM
    assert formated_variant['end_chrom'] == variant.CHROM
    assert formated_variant['pos'] == variant.POS
    assert formated_variant['end'] == variant.INFO['END']
    assert formated_variant['sv_len'] == abs(variant.INFO['SVLEN'])
    
    
    assert formated_variant['ref'] == variant.REF
    assert formated_variant['alt'] == variant.ALT[0]
    assert formated_variant['sv_type'] == 'DEL'
    assert formated_variant['case_id'] == case_id
    assert formated_variant['homozygote'] == 0
    assert formated_variant['hemizygote'] == 0

def test_format_small_ins(small_insert_variant, case_obj):
    ## GIVEN a small insertion (This means that the insertion is included in ALT field)
    variant = small_insert_variant
    case_id = case_obj['case_id']
    ## WHEN parsing the variant
    formated_variant = build_variant(
        variant=variant,
        case_obj=case_obj,
        case_id=case_id
        )
    
    ## THEN assert the sv is parsed correct
    assert formated_variant['chrom'] == variant.CHROM
    assert formated_variant['end_chrom'] == variant.CHROM
    assert formated_variant['pos'] == variant.POS
    assert formated_variant['end'] == variant.POS + abs(variant.INFO['SVLEN'])
    assert formated_variant['sv_len'] == abs(variant.INFO['SVLEN'])
    
    assert formated_variant['ref'] == variant.REF
    assert formated_variant['alt'] == variant.ALT[0]
    assert formated_variant['sv_type'] == 'INS'

def test_format_insertion(insertion_variant, case_obj):
    ## GIVEN a small insertion (This means that the insertion is included in ALT field)
    variant = insertion_variant
    case_id = case_obj['case_id']
    ## WHEN parsing the variant
    formated_variant = build_variant(
        variant=variant,
        case_obj=case_obj,
        case_id=case_id
        )
    
    ## THEN assert the sv is parsed correct
    assert formated_variant['chrom'] == variant.CHROM
    assert formated_variant['end_chrom'] == variant.CHROM
    assert formated_variant['pos'] == variant.POS
    assert formated_variant['end'] == variant.INFO['END']
    assert formated_variant['sv_len'] == 0
    
    
    assert formated_variant['ref'] == variant.REF
    assert formated_variant['alt'] == variant.ALT[0]
    assert formated_variant['sv_type'] == 'INS'
    
def test_format_dup_tandem(duptandem_variant, case_obj):
    ## GIVEN a small insertion (This means that the insertion is included in ALT field)
    variant = duptandem_variant
    case_id = case_obj['case_id']
    ## WHEN parsing the variant
    formated_variant = build_variant(
        variant=variant,
        case_obj=case_obj,
        case_id=case_id
        )
    
    ## THEN assert the sv is parsed correct
    assert formated_variant['chrom'] == variant.CHROM
    assert formated_variant['end_chrom'] == variant.CHROM
    assert formated_variant['pos'] == variant.POS
    assert formated_variant['end'] == variant.INFO['END']
    assert formated_variant['sv_len'] == abs(variant.INFO['SVLEN'])
    
    
    assert formated_variant['ref'] == variant.REF
    assert formated_variant['alt'] == variant.ALT[0]
    assert formated_variant['sv_type'] == 'DUP'
    
def test_format_translocation(translocation_variant, case_obj):
    ## GIVEN a small insertion (This means that the insertion is included in ALT field)
    variant = translocation_variant
    case_id = case_obj['case_id']
    ## WHEN parsing the variant
    formated_variant = build_variant(
        variant=variant,
        case_obj=case_obj,
        case_id=case_id
        )
    
    ## THEN assert the sv is parsed correct
    assert formated_variant['chrom'] == variant.CHROM
    assert formated_variant['end_chrom'] == '11'
    assert formated_variant['pos'] == variant.POS
    assert formated_variant['end'] == 119123896
    assert formated_variant['sv_len'] == float('inf')
    
    
    assert formated_variant['ref'] == variant.REF
    assert formated_variant['alt'] == variant.ALT[0]
    assert formated_variant['sv_type'] == 'BND'



from loqusdb.build_models import build_variant

def test_build_het_variant(het_variant, case_obj):
    variant_obj = build_variant(
        variant = het_variant,
        case_obj = case_obj
    )
    assert variant_obj['chrom'] == het_variant.CHROM
    assert variant_obj['homozygote'] == 0
    assert variant_obj['hemizygote'] == 0